#!/usr/bin/env python3

import os, sys

join = os.path.join
norm = os.path.normpath

base_dir = os.path.dirname(os.path.abspath(__file__))

lib = join(base_dir, 'lib')
sys.path.insert(0, lib)

import datetime, filecmp, hashlib, json, mimetypes, pathlib, re, shutil, ssl, string, subprocess, tempfile

import urllib.parse

import tornado.ioloop
import tornado.web

import boto3, botocore.client

from xml.dom import minidom

PORT = 8668
APPS = 'applications'

KEY_LEN = 54
HASH_SIZE = 0xFFFF

SETTINGS_JSON = 'settings.json'

LOGIC_JS = 'visual_logic.js'
LOGIC_XML = 'visual_logic.xml'
PUZZLES_DIR = 'puzzles'
LIBRARY_XML = 'my_library.xml'
BACKUP_PUZZLES_DIR = 'puzzles_backup'
BACKUP_LIBRARY_DIR = 'library_backup'
BACKUP_PUZZLES_PREFIX = 'visual_logic_'
BACKUP_LIBRARY_PREFIX = 'my_library_'

CDN_HOST = 'https://cdn.soft8soft.com/'

SERVER_MAX_BODY_SIZE = 256*1024*1024

AUTH_ADDRESS = 'auth.soft8soft.com'
AUTH_PORT = 443

ACCESS_DENIED_MSG = '<span class="red">Access is denied. Make sure that you have permissions to write into Verge3D installation directory.</span>'

class AppManagerServer(object):

    def getLocalIP(self):
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            IP = s.getsockname()[0]
        except:
            IP = ''
        finally:
            s.close()

        return IP

    def loadSettings(self):
        self.amSettings = {}

        try:
            with open(join(base_dir, SETTINGS_JSON), 'r', encoding='utf-8') as f:
                self.amSettings = json.load(f)
        except OSError:
            # using default server settings
            pass
        except json.decoder.JSONDecodeError:
            print('Settings decoding error')

        self.amSettings.setdefault('check_for_updates', True)
        self.amSettings.setdefault('external_interface', False)
        self.amSettings.setdefault('cache_max_age', 15) # minutes
        self.amSettings.setdefault('app_templates', [
            {
                "name": "Standard",
                "description": "Create a generic Puzzles-enabled app suitable for most users"
            },
            {
                "name": "Code-Based",
                "description": "Create a starter set of HTML/CSS/JavaScript files for use by web developers"
            }
        ])

        self.amSettings.setdefault('license_key', '')
        self.amSettings.setdefault('license_key_version', '')

        if 'external_address' in self.amSettings:
            del self.amSettings['external_address']

        if self.amSettings['external_interface']:
            ip = self.getLocalIP()
            if ip:
                self.amSettings['external_address'] = 'http://{0}:{1}/'.format(ip, PORT);

    def saveSettings(self):
        with open(join(base_dir, SETTINGS_JSON), 'w', encoding='utf-8') as f:
            json.dump(self.amSettings, f)
            print('App manager settings saved')

    def handleUpdate(self):
        licenseInfo = self.getLicenseInfo()

        if licenseInfo['maintenance'] and licenseInfo['releaseVersion'] != self.amSettings['license_key_version']:

            print('Updating licensing information')

            root_dir = self.getRootDir()

            for d in [APPS, 'build']:
                for v3dpath in pathlib.Path(join(root_dir, d)).rglob('v3d*.js'):
                    path = str(v3dpath)
                    self.replaceEngineKey(path, self.amSettings['license_key'])

            self.amSettings['license_key_version'] = licenseInfo['releaseVersion']
            self.saveSettings()

    def pathToUrl(self, path, quote=True, replaceBackslash=True):
        if replaceBackslash:
            path = path.replace('\\', '/')
        if quote:
            path = urllib.parse.quote(path, safe='/:')
        return path

    def replaceAmp(self, url):
        return url.replace('&', '&amp;')

    def urlBasename(self, url):
        path = urllib.parse.urlparse(url).path
        return os.path.basename(path)

    def argBool(self, argument):
        if argument.lower() in ['1', 'yes', 'y']:
            return True
        else:
            return False

    def genHTMLLink(self, path, printName=False):
        link = ''
        if printName:
            link += '<span class="icon-label-item">'
        link += '<a href="/' + self.pathToUrl(path)
        link += '" class="appbutton" title="Run application" target="_blank">'
        link += '</a>'
        if printName:
            link+='<a href="/{0}" class="button-label" target="_blank">{1}</a>&nbsp</span>'.format(
                    self.pathToUrl(path), os.path.basename(path))
        return link

    def genPlayerLink(self, path, logic_path, printName=False):
        url = '/player/player.html?load=../' + self.pathToUrl(path)
        if len(logic_path) > 0:
            url += '&logic=../' + self.pathToUrl(logic_path)

        cssclasses = 'appbutton appbutton-green'
        title = 'View scene: ' + os.path.basename(path)

        link = ''
        if printName:
            link += '<span class="icon-label-item">'
        link += '<a href="{0}" class="{1}" title="{2}" target="_blank"></a>'.format(url, cssclasses, title);
        if printName:
            link+='<a href="{0}" class="button-label" target="_blank">{1}</a>&nbsp</span>'.format(url, os.path.basename(path))
        return link

    def genHTMLLinkPuzzles(self, path, logic_path, logic_exists):
        link = '<a href="/' + self.pathToUrl(path)
        # link to logic always provided to puzzles editor
        link += '?logic=../../' + self.pathToUrl(logic_path)
        if logic_exists:
            link += '" class="logicbutton logicbutton-green" title="Edit Puzzles" target="_blank">'
        else:
            link += '" class="logicbutton logicbutton-red" title="Create Puzzles" target="_blank">'
        link += '</a>'
        return link

    def genPlayerLinkPuzzles(self, path, logic_path, logic_exists):
        link = '<a href="/player/player.html?load=../' + self.pathToUrl(path)
        # link to logic always provided to puzzles editor
        link += '&logic=../' + self.pathToUrl(logic_path)
        if logic_exists:
            link += '" class="logicbutton logicbutton-green" title="Edit Puzzles" target="_blank">'
        else:
            link += '" class="logicbutton logicbutton-red" title="Create Puzzles" target="_blank">'
        link += '</a>'
        return link

    def genBlenderLink(self, path, printName=False):

        href = 'javascript:void(0);'
        onclick = 'openFile("' + self.pathToUrl(path) + '")'
        title = 'Open file: ' + os.path.basename(path)

        link = ''
        if printName:
            link += '<span class="icon-label-item">'
        link += '<a href="{0}" onclick={1} class="blenderbutton" title="{2}"></a>'.format(href, onclick, title)
        if printName:
            link+='<a href="{0}" onclick={1} class="button-label">{2}</a>&nbsp</span>'.format(
                    href, onclick, os.path.basename(path))
        return link

    def genMaxLink(self, path, printName=False):
        href = 'javascript:void(0);'
        onclick = 'openFile("' + self.pathToUrl(path) + '")'
        title = 'Open file: ' + os.path.basename(path)

        link = ''
        if printName:
            link += '<span class="icon-label-item">'
        link += '<a href="{0}" onclick={1} class="maxbutton" title="{2}"></a>'.format(href, onclick, title)
        if printName:
            link+='<a href="{0}" onclick={1} class="button-label">{2}</a>&nbsp</span>'.format(
                    href, onclick, os.path.basename(path))
        return link

    def genMayaLink(self, path, printName=False):
        href = 'javascript:void(0);'
        onclick = 'openFile("' + self.pathToUrl(path) + '")'
        title = 'Open file: ' + os.path.basename(path)

        link = ''
        if printName:
            link += '<span class="icon-label-item">'
        link += '<a href="{0}" onclick={1} class="mayabutton" title="{2}"></a>'.format(href, onclick, title)
        if printName:
            link+='<a href="{0}" onclick={1} class="button-label">{2}</a>&nbsp</span>'.format(
                    href, onclick, os.path.basename(path))
        return link

    def genPublishLink(self, path):
        link = '<a href="javascript:void(0);" onclick=publishApp("' + self.pathToUrl(path) + '")'
        link += ' class="cloudbutton" title="Publish on the Web using Verge3D Network">'
        link += '</a>'
        return link

    def genUpdateLink(self, path, modules, files):
        # these params will be quoted during HTML form submit
        link = '<a href="javascript:void(0);" onclick=\'updateApp("{0}",[{1}],[{2}])\''.format(
            self.pathToUrl(path, quote=False),
            ','.join('"{0}"'.format(self.pathToUrl(m, quote=False)) for m in modules),
            ','.join('"{0}"'.format(self.pathToUrl(f, quote=False)) for f in files)
        )

        link += ' class="updatebutton" title="Update application">'
        link += '</a>'
        return link

    def genOpenFolderLink(self, path):

        href = 'javascript:void(0);'
        onclick = 'openFile("' + self.pathToUrl(path) + '")'

        link = '<a href="{0}" onclick={1} class="folderbutton" title="Open app folder"></a>'.format(href, onclick)
        return link

    def genLicenseInfoText(self, licenseInfo):

        if licenseInfo['type'] == 'TRIAL':
            s = '<span class="red">You are currently using the trial version of Verge3D.</span>'
            s += '<br>Please make sure to acquire a license before you start using this software in production. '
            s += '<br>Read about the benefits and limitations of Verge3D Trial <a href="https://www.soft8soft.com/verge3d-trial/" target="_blank">here</a>.'
        elif licenseInfo['maintenance']:
            s = '<span class="green">Your Verge3D for ' + licenseInfo['modPackage'].title() + ' ' + licenseInfo['type'].title() + ' License is active.</span>'
            s += '<br>You are eligible for receiving software updates and support until '
            s += licenseInfo['validUntil'].strftime('%d %b %Y') + '.'
        else:
            s = '<span class="red">Your software maintenance subscription expired on {}.</span>'.format(licenseInfo['validUntil'].date().strftime('%d %B %Y'))
            if licenseInfo['renewalGracePeriod'] > 0:
                s += '<br>It is only left {} days from your grace period to renew your subscription with 50% discount.'.format(licenseInfo['renewalGracePeriod'])
            s += '<br>Please consider renewing your Verge3D subscription.'

        return s

    def genLicenseInfoButtons(self, licenseInfo):
        s = ''

        if licenseInfo['type'] == 'TRIAL':
            s += '<button class="button license" title="Redirect to Soft8Soft licensing page" onclick="redirectPurchase()">Purchase</button>'
        elif licenseInfo['maintenance'] == False:
            s += '<button class="button license" title="Find out more about Verge3D subscription renewal" onclick="redirectRenew()">Renewal Info</button>'

        return s

    def getRootDir(self):
        return norm(join(base_dir, '..'))

    def genUpdateInfo(self, path, hjcFiles):
        """path is root relative, html/js/css files are app relative"""

        root_dir = self.getRootDir()

        MODULES = ['v3d.js', 'ie_compat.js', 'webxr-polyfill.js', 'ammo.js']

        modules = []

        needUpdate = False

        for m in MODULES:
            m_src = join(base_dir, '..', 'build', m)
            m_dst = join(root_dir, path, m)

            if os.path.exists(m_dst):
                modules.append(m)

                if not filecmp.cmp(m_src, m_dst):
                    needUpdate = True

        if not needUpdate:
            return None

        files = []

        app_name = os.path.basename(path)

        for f in hjcFiles:
            # optimization
            if os.path.basename(f) in MODULES:
                continue

            file_path = join(root_dir, path, f)
            if os.path.exists(file_path) and self.isTemplateBasedFile(file_path):
                files.append(f)

        # update media folder too
        if len(files):
            files.append('media')

        info = {
            'modules': modules,
            'files': files
        }

        return info

    def findApps(self):
        apps = []

        root_dir = self.getRootDir()

        for apppath in sorted(pathlib.Path(join(root_dir, APPS)).iterdir()):
            if apppath.is_dir():
                apps.append(self.appInfo(str(apppath)))

        return apps

    def listFilesRel(self, path, pattern, relativePath):
        paths = []

        for p in path.rglob(pattern):
            pr = str(p.relative_to(relativePath))
            # ignore hidden files
            if not os.path.basename(pr).startswith('.'):
                paths.append(pr)

        return sorted(paths)

    def appInfo(self, path):
        root_dir = self.getRootDir()

        apppath = pathlib.Path(path)

        name = os.path.basename(str(apppath))
        path_rel = str(apppath.relative_to(root_dir))

        logicXML = join(path_rel, LOGIC_XML)

        if os.path.exists(join(root_dir, logicXML)) == True:
            logicJS = join(path_rel, LOGIC_JS)
        else:
            logicJS = ''

        return {
            'name' : name,
            'title' : name.replace('_', ' ').title(),
            'path' : path_rel,
            'html' : self.listFilesRel(apppath, '*.html', root_dir),
            'blend' : self.listFilesRel(apppath, '*.blend', root_dir),
            'max' : self.listFilesRel(apppath, '*.max', root_dir),
            'maya' : (self.listFilesRel(apppath, '*.ma', root_dir) +
                      self.listFilesRel(apppath, '*.mb', root_dir)),
            'gltf' : (self.listFilesRel(apppath, '*.gltf', root_dir) +
                      self.listFilesRel(apppath, '*.glb', root_dir)),
            'logicXML': logicXML,
            'logicJS': logicJS,
            'updateInfo': self.genUpdateInfo(path_rel,
                    self.listFilesRel(apppath, '*.html', apppath) +
                    self.listFilesRel(apppath, '*.js', apppath) +
                    self.listFilesRel(apppath, '*.css', apppath))
        }

    def isPuzzlesBasedFile(self, path):
        """absolute path"""

        # latin_1 fixes issues with forbidden utf-8 characters
        with open(path, 'r', encoding='latin_1') as f:
            content = f.read()
            if '__V3D_PUZZLES__' in content:
                return True
            # COMPAT: < 2.10
            if '__VERGE3D_PLAYER__' in content:
                return True

        return False

    def isTemplateBasedFile(self, path):
        """absolute path"""

        # latin_1 fixes issues with forbidden utf-8 characters
        with open(path, 'r', encoding='latin_1') as f:
            content = f.read()
            if '__V3D_TEMPLATE__' in content:
                return True
            # COMPAT: < 2.10
            if '__VERGE3D_PLAYER__' in content:
                return True

        return False


    def checksum(self, s):
        if len(s) != KEY_LEN:
            return False

        sum = 0
        for i in range(len(s) - 4):
            sum = sum + ord(s[i])

        hash1 = sum % HASH_SIZE
        hash2 = int(s[len(s)-4 : ], 16)

        return hash1 == hash2

    def getAnonymousTrialKey(self):

        print('Receiving anonymous trial license key')

        context = ssl.create_default_context()

        # HACK: workaround for SSLCertVerificationError in python 3.7
        pyVer = sys.version_info
        if pyVer[0] >= 3 and pyVer[1] >= 7:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        # HACK: import it here to prevent strange issue with BlenderKit
        import http.client
        conn = http.client.HTTPSConnection(AUTH_ADDRESS, AUTH_PORT, context=context)

        try:
            conn.request('GET', '/get_anon_key')
            resp = conn.getresponse()
        except:
            print('GET /get_anon_key Connection refused')
            return None

        if resp.status == 200:
            key = resp.read().decode()

            self.amSettings['license_key'] = key
            self.amSettings['license_key_version'] = self.getLicenseInfo()['releaseVersion']

            self.saveSettings()
            print('Key:', key)
            return key
        else:
            print('GET /get_anon_key', resp.status, resp.reason)
            return None

    def getKeyTypeChar(self, key):
        return key[len(key)-14]

    def checkKeyModPackage(self, key):
        typechar = self.getKeyTypeChar(key)

        if self.modPackage == 'BLENDER' and typechar in ['1', '2', '3']:
            return True
        elif self.modPackage == 'MAX' and typechar in ['4', '5', '6']:
            return True
        elif self.modPackage == 'MAYA' and typechar in ['7', '8', '9']:
            return True
        else:
            return False

    def replaceEngineKey(self, path, key):
        v3dlmatch = re.compile('__V3DL__[a-fA-F0-9]{10,10}')
        v3dlnew = '__V3DL__' + key[0:10]

        lines = None

        with open(path, 'r',  encoding='utf-8') as fin:
            lines = fin.readlines()

        with open(path, 'w',  encoding='utf-8') as fout:
            for line in lines:
                fout.write(v3dlmatch.sub(v3dlnew, line))

    def getLicenseInfo(self, key=None):
        info = {
            'type': 'TRIAL',
            'modPackage': '',
            'keyExists': False,
            'keyHash': None,
            'validUntil': None,
            'maintenance': False,
            'renewalGracePeriod': 0,
            'releaseVersion': '',
            'releaseDate': None
        }

        try:
            with open(join(base_dir, '..', 'package.json'), 'r', encoding='utf-8') as f:
                package = json.load(f)
                info['releaseVersion'] = package['version']
                info['releaseDate'] = datetime.datetime.strptime(package['date'], "%Y-%m-%d")
        except OSError:
            pass

        # get from file
        if key == None:
            key = self.amSettings['license_key'].strip()

        if key and self.checksum(key):
            info['keyExists'] = True

            # generate key hash
            h = hashlib.sha1()
            h.update(key.encode())
            info['keyHash'] = h.hexdigest()

            # parse key info
            typechar = self.getKeyTypeChar(key)
            delta = 0

            if typechar == '0':
                info['type'] = 'TRIAL'
                delta = datetime.timedelta(days=30)
            else:
                if typechar == '1':
                    info['type'] = 'FREELANCE'
                    info['modPackage'] = 'BLENDER'
                if typechar == '2':
                    info['type'] = 'TEAM'
                    info['modPackage'] = 'BLENDER'
                if typechar == '3':
                    info['type'] = 'ENTERPRISE'
                    info['modPackage'] = 'BLENDER'
                if typechar == '4':
                    info['type'] = 'FREELANCE'
                    info['modPackage'] = 'MAX'
                if typechar == '5':
                    info['type'] = 'TEAM'
                    info['modPackage'] = 'MAX'
                if typechar == '6':
                    info['type'] = 'ENTERPRISE'
                    info['modPackage'] = 'MAX'
                if typechar == '7':
                    info['type'] = 'FREELANCE'
                    info['modPackage'] = 'MAYA'
                if typechar == '8':
                    info['type'] = 'TEAM'
                    info['modPackage'] = 'MAYA'
                if typechar == '9':
                    info['type'] = 'ENTERPRISE'
                    info['modPackage'] = 'MAYA'
                delta = datetime.timedelta(days=365)

                info['maintenance'] = True

            unix = int(key[len(key)-13 : len(key)-4], 16)

            dt = datetime.datetime.utcfromtimestamp(unix)

            info['validUntil'] = dt + delta

            if info['validUntil'] < info['releaseDate']:
                info['type'] = 'OUTDATED'
                info['maintenance'] = False

            now = datetime.datetime.now()
            if now > info['validUntil']:
                info['maintenance'] = False
                renewalDelta = datetime.timedelta(days=30) - (now - info['validUntil'])
                info['renewalGracePeriod'] = max(0, renewalDelta.days)

        return info

    def copyTemplateTree(self, src, dst, appName, modPackSubDir):

        names = os.listdir(src)
        errors = []

        # do not crash on already created app dir
        if not os.path.exists(dst):
            os.mkdir(dst)

        root_dir = self.getRootDir()

        for name in names:
            srcname = join(src, name)

            try:
                if os.path.isdir(srcname):
                    if name not in ['blender', 'max', 'maya']:
                        self.copyTemplateTree(srcname, join(dst, name), appName, modPackSubDir)
                    elif name == modPackSubDir:
                        # copy modelling package assets to upper dir
                        self.copyTemplateTree(srcname, dst, appName, modPackSubDir)
                else:
                    dstname = join(dst, name.replace('template', appName))
                    print(os.path.basename(srcname), '->', os.path.basename(dstname))
                    shutil.copy2(srcname, dstname)
            except OSError as why:
                errors.append((srcname, join(dst, name), str(why)))
            except shutil.Error as err:
                errors.extend(err.args[0])

        if errors:
            raise shutil.Error(errors)

    def copyAppTemplate(self, appDir, tplName, modPackage, modules):

        appName = os.path.basename(appDir)
        appTplDir = join(base_dir, 'template', 'app', tplName)

        if not os.path.exists(appTplDir):
            return

        self.copyTemplateTree(appTplDir, appDir, appName, modPackage.lower())

        htmlPath = join(appDir, appName + '.html')
        # may be omitted in user template
        if os.path.exists(htmlPath):
            self.replaceTemplateStrings(htmlPath, 'template.', appName + '.')

            compatCode = ''

            if 'ie_compat.js' in modules:
                compatCode += '<script src="ie_compat.js"></script>\n'
            if 'webxr-polyfill.js' in modules:
                compatCode += '  <script src="webxr-polyfill.js"></script>\n'
                compatCode += '  <script>var polyfill = new WebXRPolyfill();</script>\n'
            if 'ammo.js' in modules:
                compatCode += '  <script src="ammo.js"></script>\n'

            self.replaceTemplateStrings(htmlPath, '<!-- __V3D_COMPAT_MODULES__ -->', compatCode)

            licenseInfo = self.getLicenseInfo()
            metaCode = '<meta name="generator" content="Verge3D {0}">'.format(licenseInfo['releaseVersion'])
            self.replaceTemplateStrings(htmlPath, '<!-- __V3D_META__ -->', metaCode)

        jsPath = join(appDir, appName + '.js')
        # may be omitted in user template
        if os.path.exists(jsPath):
            self.replaceTemplateStrings(jsPath, 'template.gltf', appName + '.gltf')

        gltfPath = join(appDir, appName + '.gltf')
        # may be omitted in user template
        if os.path.exists(gltfPath):
            self.replaceTemplateStrings(gltfPath, 'template.bin', appName + '.bin')

        indexPath = join(appDir, 'index.html')
        # may be omitted in user template
        if os.path.exists(indexPath):
            self.replaceTemplateStrings(indexPath, 'template.html', appName + '.html')

        # copy modules from build dir
        for mod in modules:
            shutil.copy(join(base_dir, '..', 'build', mod), appDir)

    def replaceTemplateStrings(self, path, pattern, replacement, count=0):
        """
        Replace 'template' to replacement inside given file
        """

        content = None

        with open(path, 'r',  encoding='utf-8') as fin:
            content = fin.read()

        if count:
            content = content.replace(pattern, replacement, count)
        else:
            content = content.replace(pattern, replacement)

        with open(path, 'w',  encoding='utf-8') as fout:
            fout.write(content)

    def getS3Credentials(self):
        import json

        licenseInfo = self.getLicenseInfo()

        # try to receive anonymous key
        if not licenseInfo['keyExists'] and self.getAnonymousTrialKey():
            licenseInfo = self.getLicenseInfo()

        if not licenseInfo['keyExists']:
            return None

        context = ssl.create_default_context()

        # HACK: workaround for SSLCertVerificationError in python 3.7
        pyVer = sys.version_info
        if pyVer[0] >= 3 and pyVer[1] >= 7:
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE

        # HACK: import it here to prevent strange issue with BlenderKit
        import http.client
        conn = http.client.HTTPSConnection(AUTH_ADDRESS, AUTH_PORT, context=context)

        try:
            conn.request('GET', '/get_credentials?hash=%s' % licenseInfo['keyHash'])
            resp = conn.getresponse()
        except:
            print('GET /get_credentials Connection refused')
            return None

        if resp.status == 200:
            return json.loads(resp.read().decode())
        else:
            print('GET /get_credentials', resp.status, resp.reason)
            return None

    def sortHTMLs(self, files):
        """index files come first"""

        indices = []
        other = []

        for f in files:
            if 'index.htm' in f:
                indices.append(f)
            else:
                other.append(f)

        return indices + other

    def checkMainFile(self, appName, file, files):
        import difflib

        for f in files:
            if (difflib.SequenceMatcher(None, os.path.basename(f), appName).ratio() >
                    difflib.SequenceMatcher(None, os.path.basename(file), appName).ratio()):
                return False

        return True

    def findMainFile(self, appName, files):
        import difflib

        mainFile = files[0]

        for f in files:
            if (difflib.SequenceMatcher(None, os.path.basename(f), appName).ratio() >
                    difflib.SequenceMatcher(None, os.path.basename(mainFile), appName).ratio()):
                mainFile = f

        return mainFile

    class ResponseTemplates(object):
        def genTemplate(self, template, substitution):
            with open(join(base_dir, 'template', template), 'r',
                        encoding='utf-8') as f:
                tpl_in = f.read()

            server = self.settings['server']

            # always needed
            licenseInfo = server.getLicenseInfo();
            substitution['version'] = licenseInfo['releaseVersion']
            substitution['package'] = server.modPackage.title()
            substitution['releaseDate'] = licenseInfo['releaseDate'].date()
            substitution['checkForUpdates'] = str(server.amSettings['check_for_updates']).lower()

            tpl_out = string.Template(tpl_in).substitute(substitution)
            return tpl_out

        def writeStatus(self, text, backLinkURL='/', backLinkText='Return back to the main screen'):
            self.write(self.genTemplate('status.tpl', {
                'text': text,
                'backLinkURL': backLinkURL,
                'backLinkText': backLinkText
            }))

    class RootHandler(tornado.web.RequestHandler):
        def get(self):
            with open(join(base_dir, 'template', 'main.tpl'), 'r',  encoding='utf-8') as f:
                tpl_in = f.read()

            server = self.settings['server']
            server.loadSettings()

            root_dir = server.getRootDir()

            apps = server.findApps()

            apps_html = ''

            for app in apps:
                apps_html += '<tr><td>'

                apps_html += '<span class="apptext" title="{0}"><a href="/manage/?app={0}">{1}</a></span>'.format(app['name'], app['title'])

                apps_html += '</td><td class="html-gltf">'

                # NONE, HTML (copied player), GLTF (linked player)
                logicType = 'NONE'

                playerHTMLs = []

                for html in server.sortHTMLs(app['html']):
                    if server.isPuzzlesBasedFile(join(root_dir, html)):
                        logicType = 'HTML'
                        playerHTMLs.append(html)

                    apps_html += server.genHTMLLink(html)

                app_has_html = bool(len(app['html']))

                if logicType != 'HTML' and not app_has_html and len(app['gltf']):
                    logicType = 'GLTF'

                for gltf in app['gltf']:

                    if not app_has_html and server.checkMainFile(app['name'], gltf, app['gltf']):
                        logicJS = app['logicJS']
                    else:
                        logicJS = ''

                    apps_html += server.genPlayerLink(gltf, logicJS)

                apps_html += '</td><td>'

                logic_exists = bool(app['logicJS'])

                if logicType == 'HTML':
                    mainHTML = server.findMainFile(app['name'], playerHTMLs)
                    apps_html += server.genHTMLLinkPuzzles(mainHTML, app['logicXML'], logic_exists)
                elif logicType == 'GLTF':
                    mainGLTF = server.findMainFile(app['name'], app['gltf'])
                    apps_html += server.genPlayerLinkPuzzles(mainGLTF, app['logicXML'], logic_exists)

                apps_html += '</td><td class="blender-max">'

                for blend in app['blend']:
                    apps_html += server.genBlenderLink(blend)

                for maxx in app['max']:
                    apps_html += server.genMaxLink(maxx)

                for maya in app['maya']:
                    apps_html += server.genMayaLink(maya)

                apps_html += '</td><td class="short">'
                apps_html += server.genPublishLink(app['path'])

                if app['updateInfo']:
                    apps_html += '</td><td class="short">'
                    apps_html += server.genUpdateLink(app['path'],
                                                      app['updateInfo']['modules'],
                                                      app['updateInfo']['files'])
                apps_html += '</td></tr>'

            app_types_html = ''
            for at in server.amSettings['app_templates']:
                checked = 'checked' if at == server.amSettings['app_templates'][0] else ''
                br = '<br>' if at != server.amSettings['app_templates'][-1] else ''
                app_types_html += '<span title="{0}"><input type="radio" name="template_name" value="{1}" {2} class="appradio">{3}</span>{4}'.format(at['description'], at['name'], checked, at['name'], br)

            licenseInfo = server.getLicenseInfo()

            tpl_out = string.Template(tpl_in).substitute({
                'apps': apps_html,
                'appTypes': app_types_html,
                'version': licenseInfo['releaseVersion'],
                'package': server.modPackage.title(),
                'releaseDate': licenseInfo['releaseDate'].date(),
                'checkForUpdates': str(server.amSettings['check_for_updates']).lower(),
                'licenseInfoText': server.genLicenseInfoText(licenseInfo),
                'licenseInfoButtons': server.genLicenseInfoButtons(licenseInfo)
            })

            self.write(tpl_out)

    class OpenFileHandler(tornado.web.RequestHandler, ResponseTemplates):
        def get(self):
            try:
                filepath = self.get_argument('filepath')
            except tornado.web.MissingArgumentError:
                self.write('Specify file name')

            filepath = norm(join(base_dir, '..', norm(filepath)))

            if sys.platform.startswith('darwin'):
                subprocess.call(('open', '-n', filepath))
            elif os.name == 'nt':
                # ignore not-found errors
                try:
                    os.startfile(filepath)
                except:
                    pass
            elif os.name == 'posix':
                subprocess.call(('xdg-open', filepath))

            self.write('Ok')

    class CreateAppHandler(tornado.web.RequestHandler, ResponseTemplates):

        def get(self):
            try:
                name = self.get_argument('app_name')
            except tornado.web.MissingArgumentError:
                self.writeStatus('Please specify App name')
                return

            try:
                tplName = self.get_argument('template_name')
            except tornado.web.MissingArgumentError:
                self.writeStatus('Please specify App template name')
                return

            server = self.settings['server']

            try:
                copy_ie11_module = server.argBool(self.get_argument('copy_ie11_module'))
            except tornado.web.MissingArgumentError:
                copy_ie11_module = False

            try:
                copy_ar_vr_module = server.argBool(self.get_argument('copy_ar_vr_module'))
            except tornado.web.MissingArgumentError:
                copy_ar_vr_module = False

            try:
                copy_physics_module = server.argBool(self.get_argument('copy_physics_module'))
            except tornado.web.MissingArgumentError:
                copy_physics_module = False

            if name == '':
                self.writeStatus('Please specify a name for your application.')
                return

            app_dir = join(base_dir, '..', APPS, name)

            try:
                os.mkdir(app_dir)
            except FileExistsError:
                status = 'Application already exists, please remove it or try another name.'
                self.writeStatus(status)
                return
            except PermissionError:
                self.writeStatus(ACCESS_DENIED_MSG)
                return

            modules = ['v3d.js']
            if copy_ie11_module:
                modules.append('ie_compat.js')
            if copy_ar_vr_module:
                modules.append('webxr-polyfill.js')
            if copy_physics_module:
                modules.append('ammo.js')

            server.copyAppTemplate(app_dir, tplName, server.modPackage, modules)
            self.writeStatus('Project ' + name + ' created successfully.')


    class ManageAppHandler(tornado.web.RequestHandler, ResponseTemplates):

        def get(self):
            try:
                name = self.get_argument('app')
            except tornado.web.MissingArgumentError:
                self.writeStatus('Please specify App name')
                return

            server = self.settings['server']
            root_dir = server.getRootDir()

            app_dir = norm(join(base_dir, '..', APPS, name))

            info = server.appInfo(str(app_dir))

            mainPanel = ''
            sidePanel = ''

            mainPanel += '<h3>Run / View</h3>'

            logicType = 'NONE'
            playerHTMLs = []

            for html in server.sortHTMLs(info['html']):
                if server.isPuzzlesBasedFile(join(root_dir, html)):
                    logicType = 'HTML'
                    playerHTMLs.append(html)

                mainPanel += server.genHTMLLink(html, True)

            app_has_html = bool(len(info['html']))

            if logicType != 'HTML' and not app_has_html and len(info['gltf']):
                logicType = 'GLTF'

            for gltf in info['gltf']:
                if not app_has_html and server.checkMainFile(info['name'], gltf, info['gltf']):
                    logicJS = info['logicJS']
                else:
                    logicJS = ''

                mainPanel += server.genPlayerLink(gltf, logicJS, True)

            logic_exists = bool(info['logicJS'])

            if logicType == 'HTML':
                mainHTML = server.findMainFile(info['name'], playerHTMLs)
                sidePanel += server.genHTMLLinkPuzzles(mainHTML, info['logicXML'], logic_exists)
            elif logicType == 'GLTF':
                mainGLTF = server.findMainFile(info['name'], info['gltf'])
                sidePanel += server.genPlayerLinkPuzzles(mainGLTF, info['logicXML'], logic_exists)

            mainPanel += '<h3>Source Files</h3>'
            for blend in info['blend']:
                mainPanel += server.genBlenderLink(blend, True)

            for maxx in info['max']:
                mainPanel += server.genMaxLink(maxx, True)

            for maya in info['maya']:
                mainPanel += server.genMayaLink(maya, True)

            sidePanel += server.genOpenFolderLink(info['path'])

            sidePanel += server.genPublishLink(info['path'])

            if info['updateInfo']:
                sidePanel += '</td><td class="short">'
                sidePanel += server.genUpdateLink(info['path'],
                                                  info['updateInfo']['modules'],
                                                  info['updateInfo']['files'])


            self.write(self.genTemplate('manage.tpl', {
                'title': info['title'],
                'mainPanel' : mainPanel,
                'sidePanel': sidePanel
            }))

    class ProcessS3Handler(tornado.web.RequestHandler, ResponseTemplates):

        def writeS3Error(self, e):

            self.set_status(503)

            TIME_ERR_STR = ('The difference between your local system time and the server time is too large. ' +
                            'This is considered as a security issue and thus your request was denied.<br>' +
                            'Please correct your local system time and try again.')

            # NOTE: S3 upload catch and replace botocore exceptions by this one
            if isinstance(e, boto3.exceptions.S3UploadFailedError):
                if 'RequestTimeTooSkewed' in str(e):
                    self.write(TIME_ERR_STR)
                else:
                    self.write('Unknown upload error: ' + str(e))

            else:
                code = e.response['Error']['Code']

                if code == 'RequestTimeTooSkewed':
                    self.writeStatus(TIME_ERR_STR)
                else:
                    self.writeStatus('Unknown server error: ' + code)


        def uploadS3(self, s3, credentials, path):
            """relative path, returns uploaded URL"""

            server = self.settings['server']
            root_dir = server.getRootDir()

            bucket = credentials['aws_bucket']
            user_dir = credentials['aws_user_dir']

            key = user_dir + '/' + server.pathToUrl(path, quote=False)

            print('Uploading:', bucket + '/' + key)

            # NOTE: force this mime instead of new 'model/gltf+json'
            # to enable CloudFront gzip compression
            if os.path.splitext(path)[1] == '.gltf':
                mime_type = 'application/json'
            else:
                mime_type = mimetypes.guess_type(key)[0]

            if not mime_type:
                mime_type = 'application/octet-stream'

            extra_args = {
                'ContentType': mime_type,
                'ACL': 'public-read',
                'CacheControl': 'max-age=' + str(server.amSettings['cache_max_age'] * 60)
            }
            s3.upload_file(join(root_dir, path), bucket, key, ExtraArgs=extra_args)

            return CDN_HOST + key

        def uploadS3Dir(self, s3, credentials, pathDirRel):

            server = self.settings['server']

            root_dir = server.getRootDir()

            urls = []

            pathAbs = pathlib.Path(join(root_dir, norm(pathDirRel)))
            uploadList = list(pathAbs.rglob('*'))

            for p in uploadList:
                if p.is_dir():
                    continue

                # NOTE: workaround for some issue with emtpy files
                if p.lstat().st_size == 0:
                    continue

                urls.append(self.uploadS3(s3, credentials, str(p.relative_to(root_dir))))

            return urls

        @tornado.gen.coroutine
        def get(self, tail=None):
            executor = tornado.concurrent.futures.ThreadPoolExecutor(1)
            yield executor.submit(self.processingThread)

        def processingThread(self):

            server = self.settings['server']
            root_dir = server.getRootDir()

            req = self.get_argument('req')

            if req != 'status' and req != 'upload' and req != 'delete' and req != 'progress':
                self.set_status(400)
                self.write('Wrong request')
                return

            if req == 'progress':
                self.write(str(server.progress))
                return

            server.progress = 0

            print('Requesting S3 credentials')

            cred = server.getS3Credentials()
            if not cred:
                self.set_status(503)
                self.write('Failed to receive Verge3D Network credentials, possibly due to connection error.')
                return

            # Create an S3 client
            s3 = boto3.client('s3',
                aws_access_key_id = cred['aws_access_key_id'],
                aws_secret_access_key = cred['aws_secret_access_key'],
                aws_session_token = cred['aws_session_token'],
                # NOTE: fixes some issue with non-standard region
                config=botocore.client.Config(signature_version='s3v4'),
                # NOTE: fixes strange boto3 bug with non-standard region
                region_name = 'eu-central-1'
            )

            server.progress = 10

            bucket = cred['aws_bucket']
            user_dir = cred['aws_user_dir']

            if req == 'status':
                prefix = user_dir + '/'
                try:
                    resp = s3.list_objects_v2(Bucket=bucket, Prefix=prefix)
                except botocore.exceptions.ClientError as e:
                    self.writeS3Error(e)
                    return

                keys = []
                if 'Contents' in resp:
                    for obj in resp['Contents']:
                        keys.append(obj['Key'])

                    status = ''

                    status += '<span class="netcheckboxallcnt" title="Select all files">'
                    status += ' <input type="checkbox" class="netcheckboxall" id="netcheckboxall" onClick="toggleAll(this)">All'
                    status += '</span>'

                    status += '<h3>Verge3D Network directory contents</h3>'
                    for key in keys:
                        link = CDN_HOST + key
                        status += '<span title="Select file">'
                        status += ' <input type="checkbox" value="' + server.pathToUrl(key, replaceBackslash=False) + '" class="netcheckbox" id="">'
                        status += '<a href="' + server.pathToUrl(link, replaceBackslash=False) + '">' + link.replace(prefix, '') + '</a><br>'
                        status += '</span>'

                    status += '<input type="submit" value="Remove selected files" class="button netremove" onclick="removeS3Files()">'

                else:
                    status = 'Your Verge3D Network directory is empty. Try to upload something!'

                self.write(self.genTemplate('network.tpl', {'text' : status }))
                server.progress = 100
                return

            elif req == 'delete':
                keys = self.get_arguments('key')

                if not len(keys) or keys[0] == '':
                    self.writeStatus('No files selected for deletion.')
                    self.set_status(400)
                    return

                delete_keys = {'Objects' : []}

                for key in keys:
                    print('Removing key: ' + key)
                    delete_keys['Objects'].append({'Key' : key})

                try:
                    resp = s3.delete_objects(Bucket=bucket, Delete=delete_keys)
                except botocore.exceptions.ClientError as e:
                    self.writeS3Error(e)
                    return

                self.writeStatus('Delete operation complete, ' + str(len(keys)) + ' file(s) removed.',
                        '/storage/s3?req=status', 'Return back to Network directory')
                server.progress = 100
                return

            app_dir = self.get_argument('app')
            app_dir = pathlib.Path(join(root_dir, norm(app_dir.strip('/'))))

            links = []
            html_links = []

            hasEngine = False

            uploadList = list(app_dir.rglob('*'))
            progressInc = 90 / len(uploadList)
            for p in uploadList:
                if (p.is_dir() or p.suffix == '.blend' or p.suffix == '.blend1' or
                        p.suffix == '.max' or p.match('puzzles_backup/*.xml')):
                    server.progress += progressInc
                    continue

                # NOTE: workaround for some issue with emtpy files
                if p.lstat().st_size == 0:
                    server.progress += progressInc
                    continue

                try:
                    url = self.uploadS3(s3, cred, str(p.relative_to(root_dir)))
                except boto3.exceptions.S3UploadFailedError as e:
                    self.writeS3Error(e)
                    return

                links.append(url)

                if p.name == 'v3d.js':
                    hasEngine = True

                if p.suffix == '.html':
                    html_links.append(url)

                server.progress += progressInc

            info = server.appInfo(str(app_dir))

            # shared engine may be needed for some HTMLs/GLTFs
            if not hasEngine and (len(html_links) or len(info['gltf'])):
                server.progress = 93
                try:
                    links.append(self.uploadS3(s3, cred, join('build', 'v3d.js')))
                except boto3.exceptions.S3UploadFailedError as e:
                    self.writeS3Error(e)
                    return

            # pure player/gltf app
            if not len(html_links) and len(info['gltf']):

                server.progress = 95

                player_url = self.uploadS3(s3, cred, join('player', 'player.html'))

                for gltf in info['gltf']:
                    gltf_link = player_url + '?load=../' + server.pathToUrl(gltf)
                    if info['logicJS']:
                        gltf_link += '&logic=../' + server.pathToUrl(info['logicJS'])

                    html_links.append(gltf_link)
                    links.append(gltf_link)

                links.append(self.uploadS3(s3, cred, join('player', 'player.js')))
                links.append(self.uploadS3(s3, cred, join('player', 'player.css')))
                links += self.uploadS3Dir(s3, cred, join('player', 'media'))

            server.progress = 100

            if not len(links):
                self.write('No files to upload')
                return

            status = '<h3 style="margin-bottom: 12px;">You did it!</h3>'

            status += '<div class="publish-labels">Your web application has been successfully uploaded to Verge3D Network. It is time to share it via social media, a direct link or embedded into your website.</div>'

            status += '<div style="margin-bottom: 12px;"></div>'

            for link in sorted(html_links):
                status += self.genTemplate('published_app.tpl', {
                    'url' : link,
                    'shorturl' : server.urlBasename(link),
                    'text' : 'Check out this interactive web application made with Verge3D!'
                })

            status += '<div class="publish-labels">List of uploaded files:</div>'
            for link in sorted(links):
                status += '<a href="' + link + '" target="_blank">' + server.urlBasename(link) + '</a> | '

            if len(links):
                status = status[:-2]

            if server.getLicenseInfo()['type'] == 'TRIAL':
                status += '<div class="trialwarning">Please note that '
                status += 'for this trial version we promise to keep your data for 30 days only. Afterwards, uploaded files may get deleted!</div>'

            self.write(status)

    class CompressLzmaHandler(tornado.web.RequestHandler, ResponseTemplates):
        def post(self, tail=None):
            import lzma
            data = self.request.body
            data = lzma.compress(data)
            self.write(data)

    class SavePuzzlesHandler(tornado.web.RequestHandler):

        def initialize(self, isLibrary, libDelete):
            self.isLibrary = isLibrary
            self.libDelete = libDelete

        def post(self, tail=None):

            server = self.settings['server']
            root_dir = server.getRootDir()

            xmlURL = self.get_argument('xmlURL', strip=False)
            xmlURL = join(root_dir, norm(xmlURL.strip('/')))

            xmlExists = os.path.exists(xmlURL)

            # save backup
            if xmlExists:

                if self.isLibrary:
                    dirname = BACKUP_LIBRARY_DIR
                    name = BACKUP_LIBRARY_PREFIX
                else:
                    dirname = BACKUP_PUZZLES_DIR
                    name = BACKUP_PUZZLES_PREFIX

                backupDir = join(os.path.dirname(xmlURL), dirname)
                if not os.path.exists(backupDir):
                    os.mkdir(backupDir)

                now = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
                shutil.copyfile(xmlURL, join(backupDir, name + now + '.xml'))

            # save XML
            xml = self.get_argument('xml', strip=False)
            print('Saving Puzzles XML', xmlURL)
            if self.isLibrary:
                if xmlExists:
                    f = open(xmlURL, 'r', encoding='utf-8', newline='\n')
                    doc = minidom.parse(f)
                    f.close()
                    if self.libDelete:
                        categories = doc.getElementsByTagName('category')
                        for category in categories:
                            if category.hasAttribute('name'):
                                if category.getAttribute('name') == self.get_argument('codeURL', strip=False):
                                    doc.documentElement.removeChild(category)
                    else:
                        xml = minidom.parseString(xml)
                        doc.documentElement.appendChild(xml.documentElement)
                    f = open(xmlURL, 'w', encoding='utf-8', newline='\n')
                    doc.writexml(f)
                    f.close()
                else:
                    f = open(xmlURL, 'w', encoding='utf-8', newline='\n')
                    f.write('<xml>' + xml + '</xml>')
                    f.close()
            else:
                f = open(xmlURL, 'w', encoding='utf-8', newline='\n')
                f.write(xml)
                f.close()

            # save JS
            if not self.isLibrary:
                codeURL = self.get_argument('codeURL', strip=False)
                codeURL = join(root_dir, norm(codeURL.strip('/')))
                code = self.get_argument('code', strip=False)
                print('Saving Puzzles JS', codeURL)
                f = open(codeURL, 'w', encoding='utf-8', newline='\n')
                f.write(code)
                f.close()

            if xmlExists:
                return


    class EnterKeyHandler(tornado.web.RequestHandler, ResponseTemplates):

        def get(self):
            try:
                key = self.get_argument('key')
            except tornado.web.MissingArgumentError:
                self.writeStatus('Please specify a key.')
                return

            key = key.strip()

            server = self.settings['server']

            if not server.checksum(key) or not server.checkKeyModPackage(key):
                self.writeStatus('<span class="red">You entered incorrect key. ' +
                        'Please try again or contact the Verge3D support service.</span>')
                return

            licenseInfo = server.getLicenseInfo(key)

            if licenseInfo['type'] == 'OUTDATED':
                self.writeStatus(
                    '<span class="red">Your software maintenance subscription expired on {}.</span>'
                    .format(licenseInfo['validUntil'].date().strftime('%d %B %Y')) +
                    '<br>Please consider <a href="https://www.soft8soft.com/verge3d-subscription-renewal/">' +
                    'renewing your subscription</a> or stick to Verge3D version ' +
                    'issued prior the expiration date.')
                return

            try:
                server.amSettings['license_key'] = key
                server.amSettings['license_key_version'] = licenseInfo['releaseVersion']
                server.saveSettings()
            except PermissionError:
                self.writeStatus(ACCESS_DENIED_MSG)
                return

            root_dir = server.getRootDir()

            for d in [APPS, 'build']:
                for v3dpath in pathlib.Path(join(root_dir, d)).rglob('v3d*.js'):
                    path = str(v3dpath)
                    server.replaceEngineKey(path, key)

            s = 'Your Verge3D for ' + licenseInfo['modPackage'].title() + ' ' + licenseInfo['type'].title() + ' License has been activated. Thank you for choosing Verge3D!'

            if licenseInfo['type'] == 'TRIAL':
                s += '<br>Please note that the key you just entered is intended for preview purposes only. '
                s += 'The key will expire after 30 day period. '
                s += 'Make sure you purchase a license for use in production.'

            self.writeStatus(s)

    class UpdateAppHandler(tornado.web.RequestHandler, ResponseTemplates):

        def get(self):
            try:
                app_dir = self.get_argument('app')
            except tornado.web.MissingArgumentError:
                self.writeStatus('Please specify application filepath')
                self.set_status(400)
                return

            modules = self.get_arguments('module')
            files = self.get_arguments('file')

            if (not len(modules) or modules[0] == '') and (not len(files) or files[0] == ''):
                self.writeStatus('No files selected for updating.')
                self.set_status(400)
                return

            server = self.settings['server']
            root_dir = server.getRootDir()

            path = join(root_dir, norm(app_dir))

            if len(files):
                tpl_name = 'Standard'
            else:
                tpl_name = 'Code-Based'

            templateAppName = os.path.basename(path)

            # NOTE: use similar-to-app name HTML name instead of updated app name
            # this resolves updating issues with renamed apps
            templateFiles = list(filter(lambda f: os.path.isfile(join(path, f)) and
                                        server.isTemplateBasedFile(join(path, f)), files))
            if len(templateFiles):
                templateAppName = os.path.splitext(server.findMainFile(os.path.basename(path), templateFiles))[0]

            with tempfile.TemporaryDirectory() as td:
                tmp_app_dir = join(td, templateAppName)
                os.mkdir(tmp_app_dir)

                # this method will fail first if the verge3d dir is read-only
                try:
                    server.copyAppTemplate(tmp_app_dir, tpl_name, server.modPackage, modules)
                except PermissionError:
                    self.writeStatus(ACCESS_DENIED_MSG)
                    return

                for m in modules:
                    print('Updating module {0}'.format(m))
                    shutil.copyfile(join(tmp_app_dir, m), join(path, m))

                for f in files:

                    if f == 'media':
                        print('Updating media dir')

                        media_dir_src = join(tmp_app_dir, 'media')
                        media_dir_dst = join(path, 'media')

                        for item in os.listdir(media_dir_src):
                           shutil.copy(join(media_dir_src, item), media_dir_dst)
                    else:
                        print('Updating file {0}'.format(f))
                        ext = os.path.splitext(f)[1]
                        shutil.copyfile(join(tmp_app_dir, os.path.basename(tmp_app_dir)+ext), join(path, f))

            self.writeStatus('The app has been successfully updated.')

    class SettingsHandler(tornado.web.RequestHandler, ResponseTemplates):
        def set_default_headers(self):
            # Fix issues with browser back-forward buttons
            self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate')
            self.set_header('Pragma', 'no-cache')
            self.set_header('Expires', '0')

        def get(self):
            server = self.settings['server']
            self.write(self.genTemplate('settings.tpl', {
                'json' : json.dumps(server.amSettings)
            }))

    class SaveSettingsHandler(tornado.web.RequestHandler, ResponseTemplates):
        def post(self, tail=None):
            data = self.request.body
            try:
                with open(join(base_dir, SETTINGS_JSON), 'wb') as f:
                    f.write(data)
                    server = self.settings['server']
                    self.write('ok')
                server.loadSettings()
            except PermissionError:
                self.write('error')
                return

    class StaticHandler(tornado.web.StaticFileHandler):
        # disable cache
        def set_extra_headers(self, path):
            # https://github.com/e0ne/BlogSamples/blob/master/tornado-nocache/handlers.py
            self.set_header('Cache-Control', 'no-cache, no-store, must-revalidate, max-age=0')
            self.set_header('Expires', '0')
            now = datetime.datetime.now()
            expiration = datetime.datetime(now.year-1, now.month, now.day)
            self.set_header('Last-Modified', expiration)

    def start(self, modPackage):

        print('Starting App Manager server (port {0}, package {1})'.format(str(PORT), modPackage))

        self.modPackage = modPackage
        self.progress = 0;

        self.loadSettings()
        self.handleUpdate()

        # check if user-created puzzles library exists - if not create one
        libXMLPath = join(self.getRootDir(), PUZZLES_DIR, LIBRARY_XML)
        if not os.path.exists(libXMLPath):
            try:
                print('Puzzles library file not found, creating one')
                f = open(libXMLPath, 'w', encoding='utf-8', newline='\n')
                f.write('<?xml version="1.0" ?><xml/>')
                f.close()
            except OSError:
                print('Access denied: Puzzles library file was not created')
                pass

        app = self.createTornadoApp()
        try:
            address = '0.0.0.0' if self.amSettings['external_interface'] else 'localhost'
            app.listen(PORT, address=address, max_body_size=SERVER_MAX_BODY_SIZE)
        except OSError:
            print('Address already in use, exiting')
            return
        else:
            tornado.ioloop.IOLoop.current().start()

    def createTornadoApp(self):
        root = os.path.join(os.path.dirname(__file__), '..')

        return tornado.web.Application([
            (r'/?$', self.RootHandler),
            (r'/create/?$', self.CreateAppHandler),
            (r'/manage/?$', self.ManageAppHandler),
            (r'/open/?$', self.OpenFileHandler),
            (r'/storage/s3/?$', self.ProcessS3Handler),
            (r'/storage/lzma/?$', self.CompressLzmaHandler),
            (r'/storage/xml/?$', self.SavePuzzlesHandler, {'isLibrary': False, 'libDelete': False}),
            (r'/storage/xml/library/?$', self.SavePuzzlesHandler, {'isLibrary': True, 'libDelete': False}),
            (r'/storage/xml/libdelete/?$', self.SavePuzzlesHandler, {'isLibrary': True, 'libDelete': True}),
            (r'/enterkey/?$', self.EnterKeyHandler),
            (r'/update_app/?$', self.UpdateAppHandler),
            (r'/settings/?$', self.SettingsHandler),
            (r'/settings/save?$', self.SaveSettingsHandler),
            (r'/(.*)$', self.StaticHandler, {'path': root, 'default_filename': 'index.html'}),
        ], server=self)


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1].upper() in ['BLENDER', 'MAX', 'MAYA']:
        modPackage = sys.argv[1].upper()
    else:
        modPackage = 'NONE'

    if len(sys.argv) > 2 and sys.argv[2].upper() == 'RUN_BROWSER':
        import webbrowser
        webbrowser.open('http://localhost:{}/'.format(PORT))

    server = AppManagerServer()
    server.start(modPackage)
