
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

from pluginUtils.log import printLog

import maya.cmds as cmds
import maya.OpenMaya as OpenMaya

import mayaUtils

class Collector():

    defaultCameras = ['|top', '|front', '|side', '|persp']

    def __init__(self, exportSettings):

        self.exportSettings = exportSettings;

        self.nodes = []
        self.topLevelNodes = []
        self.meshes = []
        self.cameras = []
        self.lights = []
        self.skyDomeLights = []
        self.materials = []
        self.textures = []

        self.convertedMeshes = []

    def collect(self):

        nodes = cmds.ls(assemblies=True, long=True)

        for node in nodes:
            self.collectNode(node, True)

        # TODO: put active camera (if any) first

        for node in self.meshes:
            shadingGrps = cmds.listConnections(node, type='shadingEngine')
            for mat in cmds.ls(cmds.listConnections(shadingGrps), materials=True):
                self.collectMat(mat, node)

        # collect textures
        for mat in self.materials:
            matType = cmds.objectType(mat)

            if matType in ['phong', 'lambert', 'blinn']:
                tex = mayaUtils.extractTexture(mat, 'color')
                if tex:
                    self.textures.append(tex)

            elif matType == 'aiStandardSurface' or matType == 'standardSurface':
                tex = mayaUtils.extractTexture(mat, 'baseColor')
                if tex:
                    self.textures.append(tex)

                tex = mayaUtils.extractTexture(mat, 'normalCamera', 'bumpValue')
                if tex:
                    self.textures.append(tex)

                tex = mayaUtils.extractTexture(mat, 'emissionColor')
                if tex:
                    self.textures.append(tex)

            elif matType == 'StingrayPBS':
                tex = mayaUtils.extractTexture(mat, 'TEX_color_map')
                if tex:
                    self.textures.append(tex)

        if len(self.skyDomeLights):
            # NOTE: get the first one
            envTex = mayaUtils.extractTexture(self.skyDomeLights[0], 'color')
            if envTex:
                self.textures.append(envTex)

        if len(self.cameras) == 0:
            node = '|persp'

            childNodes = cmds.listRelatives(node, children=True, fullPath=True)
            if childNodes:
                for child in childNodes:
                    if mayaUtils.isCamera(child):
                        self.cameras.append(child)
                        self.nodes.append(node)
                        self.topLevelNodes.append(node)

    def collectNode(self, node, topLevel):
        if node in self.defaultCameras:
            return

        node = self.convertNode(node)

        self.nodes.append(node)

        if topLevel:
            self.topLevelNodes.append(node)

        childNodes = cmds.listRelatives(node, children=True, fullPath=True)
        if childNodes:
            for child in childNodes:
                if mayaUtils.isMesh(child):
                    self.meshes.append(child)
                elif mayaUtils.isCamera(child):
                    self.cameras.append(child)
                elif mayaUtils.isLight(child):
                    self.lights.append(child)
                elif mayaUtils.isSkyDomeLight(child):
                    self.skyDomeLights.append(child)
                elif mayaUtils.isTransformNode(child):
                    self.collectNode(child, False)

    def convertNode(self, node):
        """Returns new node if conversion needed"""

        childNodes = cmds.listRelatives(node, children=True, fullPath=True)
        if childNodes:
            shape = childNodes[0]

            if mayaUtils.isNurbsSurface(shape):
                surfName = mayaUtils.getName(node)
                printLog('INFO', 'Converting {} surface to poly mesh'.format(surfName))
                nodeConv = cmds.nurbsToPoly(node,
                                            name = surfName + 'Converted',
                                            constructionHistory = True,
                                            format = 1,
                                            polygonType = 1,
                                            fractionalTolerance = 0.01,
                                            minEdgeLength = 0.001,
                                            delta = 0.1,
                                            useSurfaceShader = True)[0]

                # restore parent link
                nodeParent = cmds.listRelatives(node, parent=True, fullPath=True)
                if nodeParent:
                    nodeConv = cmds.ls(cmds.parent(nodeConv, nodeParent[0]), long=True)[0]

                self.convertedMeshes.append(nodeConv)

                node = nodeConv

        return node


    def collectMat(self, mat, node):
        if mat and mat not in self.materials:
            self.materials.append(mat)


    def cleanup(self):
        for node in self.convertedMeshes:
            cmds.delete(node)
