
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

import math

from pluginUtils.log import printLog

import maya.cmds as cmds

AA_METHODS_MAP = [
    'AUTO',
    'MSAA4',
    'MSAA8',
    'MSAA16',
    'FXAA'
]

ANIM_LOOP_MAP = [
    'REPEAT',
    'PING_PONG',
    'ONCE'
]

CAMERA_CONTROLS_MAP = [
    'ORBIT',
    'FLYING',
    'FIRST_PERSON',
    'NONE'
]

def initSettingsNode(nodeType):

    cmds.addExtension(nodeType=nodeType, longName='copyright', dataType='string',
                      niceName='Scene Copyright')
    cmds.addExtension(nodeType=nodeType, longName='lzmaEnabled', attributeType='bool',
                      niceName='Enable LZMA Compression')
    cmds.addExtension(nodeType=nodeType, longName='aaMethod', attributeType='enum',
                      enumName='Auto:MSAA 4x:MSAA 8x:MSAA 16x:FXAA', niceName='Anti-Aliasing')
    cmds.addExtension(nodeType=nodeType, longName='useHDR', attributeType='bool',
                      niceName='Use HDR Rendering')
    cmds.addExtension(nodeType=nodeType, longName='optimizeAttrs', attributeType='bool', defaultValue=True,
                      niceName='Optimize Mesh Attributes')

    cmds.addExtension(nodeType=nodeType, longName='outlineEnabled', attributeType='bool',
                      niceName='Enabled')
    cmds.addExtension(nodeType=nodeType, longName='edgeStrength', attributeType='float',
                      defaultValue=3, minValue=0, maxValue=10)
    cmds.addExtension(nodeType=nodeType, longName='edgeGlow', attributeType='float',
                      defaultValue=0, minValue=0, maxValue=10)
    cmds.addExtension(nodeType=nodeType, longName='edgeThickness', attributeType='float',
                      defaultValue=1, minValue=0, maxValue=10)
    cmds.addExtension(nodeType=nodeType, longName='pulsePeriod', attributeType='float', defaultValue=0)
    cmds.addExtension(nodeType=nodeType, longName='visibleEdgeColor', attributeType='float3', usedAsColor=True)
    cmds.addExtension(nodeType=nodeType, longName='colorR', attributeType='float', parent='visibleEdgeColor')
    cmds.addExtension(nodeType=nodeType, longName='colorG', attributeType='float', parent='visibleEdgeColor')
    cmds.addExtension(nodeType=nodeType, longName='colorB', attributeType='float', parent='visibleEdgeColor')
    cmds.addExtension(nodeType=nodeType, longName='hiddenEdgeColor', attributeType='float3', usedAsColor=True)
    cmds.addExtension(nodeType=nodeType, longName='colorR', attributeType='float', parent='hiddenEdgeColor')
    cmds.addExtension(nodeType=nodeType, longName='colorG', attributeType='float', parent='hiddenEdgeColor')
    cmds.addExtension(nodeType=nodeType, longName='colorB', attributeType='float', parent='hiddenEdgeColor')

    cmds.addExtension(nodeType=nodeType, longName='animExport', attributeType='bool', defaultValue=True,
                      niceName='Export Animations')
    cmds.addExtension(nodeType=nodeType, longName='animUsePlaybackRange', attributeType='bool',
                      niceName='Export Within Playback Range')
    cmds.addExtension(nodeType=nodeType, longName='animStartWithZero', attributeType='bool', defaultValue=True,
                      niceName='Keyframes Start with 0')


def createSettingsNode():
    """
    instantiate v3dSettings node
    """

    if len(cmds.ls(type='v3dSettings')) == 0:
        cmds.createNode('v3dSettings', shared=True, name='v3dSettings', skipSelect=True)

def removeSettingsNode():
    """
    remove v3dSettings node if any
    """

    for node in cmds.ls(type='v3dSettings'):
        cmds.delete(node)

def parseSettings(output={}):

    defaultSettings = {
        'copyright': '',
        'lzmaEnabled': False,
        'aaMethod': 'AUTO',
        'useHDR': False,
        'optimizeAttrs': True,
        'outlineEnabled': False,
        'edgeStrength': 3,
        'edgeGlow': 0,
        'edgeThickness': 1,
        'pulsePeriod': 0,
        'visibleEdgeColor': [1, 1, 1],
        'hiddenEdgeColor': [0.1, 0.1, 0.1],
        'animExport': True,
        'animUsePlaybackRange': False,
        'animStartWithZero': True
    }

    if len(cmds.ls(type='v3dSettings')) == 0:
        printLog('INFO', 'Using default Verge3D settings')
        output.update(defaultSettings)
        return

    for name in defaultSettings:
        value = cmds.getAttr('v3dSettings.' + name)

        if name == 'aaMethod':
            value = AA_METHODS_MAP[value]

        # color attributes return one-element list with tuple
        elif cmds.attributeQuery(name, node='v3dSettings', usedAsColor=True):
            value = list(value[0])

        output[name] = value

def assignNodeSettings():
    assignAnimationSettings()
    assignCameraSettings()

def assignAnimationSettings():
    cmds.addExtension(nodeType='transform', longName='v3d', attributeType='compound',
                      numberOfChildren=8, niceName='Verge3D')

    cmds.addExtension(nodeType='transform', longName='animAuto', attributeType='bool',
                      defaultValue=True, parent='v3d', niceName='Auto Start')

    cmds.addExtension(nodeType='transform', longName='animLoop', attributeType='enum',
                      enumName='Repeat:Ping Pong:Once', parent='v3d', niceName='Loop Mode')

    cmds.addExtension(nodeType='transform', longName='animRepeatInfinite', attributeType='bool',
                      defaultValue=True, parent='v3d', niceName='Repeat Infinitely')
    cmds.addExtension(nodeType='transform', longName='animRepeatCount', attributeType='float',
                      defaultValue=1, minValue=0, softMaxValue=1000, parent='v3d', niceName='Repeat Count')
    cmds.addExtension(nodeType='transform', longName='animOffset', attributeType='float',
                      defaultValue=0, softMinValue=-1000, softMaxValue=1000, parent='v3d', niceName='Offset')

    cmds.addExtension(nodeType='transform', longName='animUseCustomRange', attributeType='bool',
                      defaultValue=False, parent='v3d', niceName='Custom Frame Range')
    cmds.addExtension(nodeType='transform', longName='animCustomRangeFrom', attributeType='float',
                      defaultValue=0, softMinValue=-1000, softMaxValue=1000, parent='v3d', niceName='From')
    cmds.addExtension(nodeType='transform', longName='animCustomRangeTo', attributeType='float',
                      defaultValue=100, softMinValue=-1000, softMaxValue=1000, parent='v3d', niceName='To')

def parseAnimationSettings(mNode):

    output = {}

    # use None because we might want some default values later

    defaultSettings = {
        'animAuto': None,
        'animLoop': None,
        'animRepeatInfinite': None,
        'animRepeatCount': None,
        'animOffset': None,
        'animUseCustomRange': None,
        'animCustomRangeFrom': None,
        'animCustomRangeTo': None
    }

    for name in defaultSettings:
        value = cmds.getAttr(mNode + '.v3d.' + name)

        if name == 'animLoop':
            value = ANIM_LOOP_MAP[value]

        output[name] = value

    return output


def assignCameraSettings():
    PI = math.pi
    PI2 = 2 * math.pi

    cmds.addExtension(nodeType='camera', longName='v3d', attributeType='compound',
                      numberOfChildren=10, niceName='Verge3D')

    cmds.addExtension(nodeType='camera', longName='controls', attributeType='enum',
                      enumName='Orbit:Flying:First-Person:Disabled', parent='v3d',
                      niceName='Controls')

    cmds.addExtension(nodeType='camera', longName='panningEnabled', attributeType='bool',
                      defaultValue=True, parent='v3d')

    cmds.addExtension(nodeType='camera', longName='rotateSpeed', attributeType='float',
                      defaultValue=1, minValue=0, softMaxValue=10, parent='v3d')
    cmds.addExtension(nodeType='camera', longName='moveSpeed', attributeType='float',
                      defaultValue=1, minValue=0, softMaxValue=10, parent='v3d')

    cmds.addExtension(nodeType='camera', longName='minDist', attributeType='float',
                      defaultValue=0, minValue=0, softMaxValue=100, parent='v3d')
    cmds.addExtension(nodeType='camera', longName='maxDist', attributeType='float',
                      defaultValue=1000, minValue=0, softMaxValue=1000, parent='v3d')
    cmds.addExtension(nodeType='camera', longName='minAngle', attributeType='doubleAngle',
                      defaultValue=0, minValue=0, maxValue=PI2, parent='v3d')
    cmds.addExtension(nodeType='camera', longName='maxAngle', attributeType='doubleAngle',
                      defaultValue=PI, minValue=0, maxValue=PI2, parent='v3d')
    cmds.addExtension(nodeType='camera', longName='minAzimuthAngle', attributeType='doubleAngle',
                      defaultValue=0, minValue=0, maxValue=PI2, parent='v3d')
    cmds.addExtension(nodeType='camera', longName='maxAzimuthAngle', attributeType='doubleAngle',
                      defaultValue=PI2, minValue=0, maxValue=PI2, parent='v3d')

def parseCameraSettings(mNode):

    output = {}

    defaultSettings = {
        'panningEnabled': None,
        'rotateSpeed': None,
        'moveSpeed': None,
        'controls': None,
        'minDist': None,
        'maxDist': None,
        'minAngle': None,
        'maxAngle': None,
        'minAzimuthAngle': None,
        'maxAzimuthAngle': None
    }

    for name in defaultSettings:
        value = cmds.getAttr(mNode + '.v3d.' + name)

        if name == 'controls':
            value = CAMERA_CONTROLS_MAP[value]
        elif name in ['minAngle', 'maxAngle', 'minAzimuthAngle', 'maxAzimuthAngle']:
            value = math.radians(value)

        output[name] = value

    return output
