from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

import json, math, os, shutil, struct, sys

from profilehooks import profile

import maya.cmds as cmds
import maya.OpenMaya as OpenMaya

join = os.path.join

from pluginUtils.log import printLog
from pluginUtils.manager import AppManagerConn
from pluginUtils import gltf

from collect import Collector
from mayaUtils import (getName, isMesh, isCamera, isLight, isTransformNode,
                       isPbrMaterial, getFPS, extractMeshData, extractTexture,
                       extractTexturePath, getRotationQuat, extractAnimData,
                       getInputNode)

import mayaUtils
import customAttrs

WORLD_NODE_MAT_NAME = 'Verge3D_Environment'
CAM_ANGLE_EPSILON = math.pi / 180


try:
    from PySide.QtGui import QImage, QColor, qRed, qGreen, qBlue, QImageWriter
except ImportError:
    from PySide2.QtGui import QImage, QColor, qRed, qGreen, qBlue, QImageWriter

class GLTFExporter(object):
    def __init__(self):
        self.output = {}
        self.binary = bytearray()

    def generateAsset(self):
        asset = {}

        asset['version'] = '2.0'
        asset['generator'] = 'Soft8Soft Verge3D for Maya plug-in'

        copyright = self.exportSettings['copyright']

        if copyright != '':
            asset['copyright'] = copyright

        self.output['asset'] = asset


    def createAnimation(self, collector, mNode):
        channels = []
        samplers = []

        animUsePlaybackRange = self.exportSettings['animUsePlaybackRange']
        animStartWithZero = self.exportSettings['animStartWithZero']

        samplerIdx = 0

        if cmds.keyframe(mNode, attribute='translate', query=True, keyframeCount=True):
            keys, values, interp = extractAnimData(mNode, 'translation')

            channels.append(gltf.createAnimChannel(samplerIdx,
                    gltf.getNodeIndex(self.output, mNode), 'translation'))
            samplers.append(gltf.createAnimSampler(self.output, self.binary,
                    keys, values, 3, interp))
            samplerIdx += 1

        if cmds.keyframe(mNode, attribute='rotate', query=True, keyframeCount=True):
            keys, values, interp = extractAnimData(mNode, 'rotation')

            channels.append(gltf.createAnimChannel(samplerIdx,
                    gltf.getNodeIndex(self.output, mNode), 'rotation'))
            samplers.append(gltf.createAnimSampler(self.output, self.binary,
                    keys, values, 4, interp))
            samplerIdx += 1

        if cmds.keyframe(mNode, attribute='scale', query=True, keyframeCount=True):
            keys, values, interp = extractAnimData(mNode, 'scale')

            channels.append(gltf.createAnimChannel(samplerIdx,
                    gltf.getNodeIndex(self.output, mNode), 'scale'))
            samplers.append(gltf.createAnimSampler(self.output, self.binary,
                    keys, values, 3, interp))
            samplerIdx += 1

        if len(channels) and len(samplers):
            animation = {
                'name': getName(mNode),
                'channels': channels,
                'samplers': samplers
            }

            animSettings = customAttrs.parseAnimationSettings(mNode)

            v3dExt = gltf.appendExtension(self.output, 'S8S_v3d_animation_data', animation)

            v3dExt['auto'] = animSettings['animAuto']
            v3dExt['loop'] = animSettings['animLoop']

            v3dExt['repeatInfinite'] = animSettings['animRepeatInfinite']
            v3dExt['repeatCount'] = animSettings['animRepeatCount']
            v3dExt['offset'] = animSettings['animOffset']
        else:
            animation = None

        return animation

    def generateAnimations(self, collector):

        if not self.exportSettings['animExport']:
            return

        animations = []

        for mNode in collector.nodes:

            anim = self.createAnimation(collector, mNode)
            if anim:
                animations.append(anim)

        if len(animations) > 0:
            self.output['animations'] = animations

    def generateBuffers(self):
        byteLength = len(self.binary)

        if byteLength > 0:
            buffer = {
                'byteLength' : byteLength
            }

            if self.exportSettings['format'] == 'ASCII':
                buffer['uri'] = self.exportSettings['binaryfilename']

            self.output['buffers'] = [buffer]


    def createCamera(self, mCamera):
        camera = {
            'name': getName(mCamera),
            'id': mCamera
        }

        camera['znear'] = cmds.camera(mCamera, query=True, nearClipPlane=True)
        camera['zfar'] = cmds.camera(mCamera, query=True, farClipPlane=True)

        if cmds.camera(mCamera, query=True, orthographic=True):
            camera['type'] = 'orthographic'
            camera['orthographic'] = {}

            camera['orthographic']['xmag'] = cmds.camera(mCamera, query=True, orthographicWidth=True)
            camera['orthographic']['ymag'] = camera['orthographic']['xmag']
        else:
            camera['type'] = 'perspective'
            camera['perspective'] = {}
            camera['perspective']['aspectRatio'] = cmds.camera(mCamera, query=True, aspectRatio=True)
            camera['perspective']['yfov'] = math.radians(cmds.camera(mCamera, query=True, verticalFieldOfView=True))

        camSettings = customAttrs.parseCameraSettings(mCamera)

        v3dExt = gltf.appendExtension(self.output, 'S8S_v3d_camera_data', camera)

        v3dExt['viewportFitType'] = 'VERTICAL'
        #v3dExt['viewportFitInitialAspect'] = aspectRatio

        v3dExt['enablePan'] = camSettings['panningEnabled']
        v3dExt['rotateSpeed'] = camSettings['rotateSpeed']
        v3dExt['moveSpeed'] = camSettings['moveSpeed']

        v3dExt['controls'] = camSettings['controls']

        if camSettings['controls'] == 'ORBIT':
            v3dExt['orbitTarget'] = [0, 0, 0]

            v3dExt['orbitMinDistance'] = camSettings['minDist']
            v3dExt['orbitMaxDistance'] = camSettings['maxDist']
            v3dExt['orbitMinPolarAngle'] = camSettings['minAngle']
            v3dExt['orbitMaxPolarAngle'] = camSettings['maxAngle']

            minAzAngle = camSettings['minAzimuthAngle']
            maxAzAngle = camSettings['maxAzimuthAngle']

            # export only when needed
            if abs(2 * math.pi - (maxAzAngle - minAzAngle)) > CAM_ANGLE_EPSILON:
                v3dExt['orbitMinAzimuthAngle'] = minAzAngle
                v3dExt['orbitMaxAzimuthAngle'] = maxAzAngle

        return camera

    def generateCameras(self, collector):
        cameras = []

        mCameras = collector.cameras

        if len(mCameras):
            for mCamera in mCameras:
                cameras.append(self.createCamera(mCamera))

            self.output['cameras'] = cameras

    def createDefaultLight(self):
        printLog('INFO', 'Generating default light')

        light = {
            'name': '__DEFAULT__',
            'profile': 'maya',
            'color': [1,1,1],
            'intensity': 1,
            'type': 'directional',
        }

        return light

    def createLight(self, mLight):

        light = {}
        light['name'] = getName(mLight)
        light['id'] = mLight
        light['profile'] = 'maya'

        lightType = cmds.objectType(mLight)

        if lightType == 'ambientLight':
            light['type'] = 'ambient'
            queryFun = cmds.ambientLight
        elif lightType == 'directionalLight':
            light['type'] = 'directional'
            queryFun = cmds.directionalLight
        elif lightType == 'pointLight':
            light['type'] = 'point'
            queryFun = cmds.pointLight
        elif lightType == 'spotLight':
            light['type'] = 'spot'
            queryFun = cmds.spotLight
        else:
            printLog('ERROR', 'Unknown light type: ' + lightType)

        light['color'] = queryFun(mLight, query=True, rgb=True)
        light['intensity'] = queryFun(mLight, query=True, intensity=True)

        # NOTE: there is no "photometry" in Maya <2020 viewport
        if light['type'] != 'ambient' and int(cmds.about(apiVersion=True)) < 20200000:
            light['intensity'] *= math.pi

        if (light['type'] == 'point' or light['type'] == 'spot'):

            #light['distance'] =
            light['decay'] = queryFun(mLight, query=True, decayRate=True)

        if light['type'] == 'spot':
            light['angle'] = math.radians(queryFun(mLight, query=True, coneAngle=True) / 2)
            light['penumbra'] = queryFun(mLight, query=True, penumbra=True)

        return light

    def generateLights(self, collector):

        lights = []

        mLights = collector.lights

        if len(mLights):
            for mLight in mLights:
                lights.append(self.createLight(mLight))
        elif len(collector.skyDomeLights) == 0:
            lights.append(self.createDefaultLight())

        gltf.appendExtension(self.output, 'S8S_v3d_data', self.output, {'lights': lights})

    def createDefaultLightNode(self, cameraNode):
        node = {}

        node['name'] = '__DEFAULT_LIGHT__'

        node['translation'] = cameraNode['translation'][:]

        light = gltf.getLightIndex(self.output, '__DEFAULT__')
        if light >= 0:
            gltf.appendExtension(self.output, 'S8S_v3d_node_data', node, {'light': light})

        return node

    def createImage(self, filePath, qimage=None):
        fileName = os.path.basename(filePath)
        base, ext = os.path.splitext(filePath)

        mimeSuffix = ext.lower()[1:]

        if mimeSuffix == 'jpg':
            mimeSuffix = 'jpeg'

        if qimage:
            writer = QImageWriter(filePath, mimeSuffix);
            writer.write(qimage);
            del writer

        image = {
            'id': filePath,
            'mimeType': gltf.imageMimeType(filePath)
        }

        if self.exportSettings['format'] == 'ASCII':
            if not qimage:
                destPath = join(self.exportSettings['filedirectory'], os.path.basename(filePath))
                if filePath != destPath:
                    shutil.copy(filePath, self.exportSettings['filedirectory'])
            image['uri'] = fileName
        else:
            with open(filePath, 'rb') as f:
                imgBytes = f.read()

            if qimage:
                os.remove(filePath)

            bufferView = gltf.generateBufferView(self.output, self.binary, imgBytes, 0, 0)
            image['bufferView'] = bufferView

        return image

    def createMetallicRoughnessMap(self, metalMap, roughMap, name):

        metal = QImage(metalMap)
        rough = QImage(roughMap)

        metalPixel = QColor()

        metal = metal.convertToFormat(QImage.Format_RGB32);
        rough = rough.convertToFormat(QImage.Format_RGB32);

        metalUCharPtr = metal.bits()
        roughUCharPtr = rough.bits()

        if (not metal.width() == rough.width()
                or not metal.height() == rough.height()):
            raise RuntimeError('Error processing material: {}. Metallic map and roughness map must have same dimensions.'.format(name))

        width = metal.width();
        height = metal.height();

        i = 0
        for y in range(height):
            for x in range(width):
                metal_color = struct.unpack('I', metalUCharPtr[i:i+4])[0]
                rough_color = struct.unpack('I', roughUCharPtr[i:i+4])[0]
                metalPixel.setRgb(0, qGreen(rough_color), qBlue(metal_color))
                metalUCharPtr[i:i+4] = struct.pack('I', metalPixel.rgb())
                i+=4

        output = self.exportSettings['filedirectory'] + name + '_rough_metal.jpg'
        return output, metal

    def generateImages(self, collector):

        images = []

        for mTex in collector.textures:
            texPath = extractTexturePath(mTex)
            images.append(self.createImage(texPath))

        if len(images) > 0:
            self.output['images'] = images

    def createTextureFromImage(self, image,
                               magFilter=gltf.WEBGL_FILTERS['LINEAR'],
                               wrapS=gltf.WEBGL_WRAPPINGS['REPEAT'],
                               wrapT=gltf.WEBGL_WRAPPINGS['REPEAT']):

        texture = {
            'name': os.path.basename('image'),
            'id': image['id'],
            'sampler': gltf.createSampler(self.output, magFilter, wrapS, wrapT),
            'source': gltf.getImageIndex(self.output, image['id'])
        }

        return texture

    def generateTextures(self, collector):
        textures = []

        for mTex in collector.textures:

            texture = {}

            texture['name'] = getName(mTex)
            texture['id'] = mTex

            magFilter = gltf.WEBGL_FILTERS['LINEAR']

            wrapS = gltf.WEBGL_WRAPPINGS['REPEAT']
            wrapT = gltf.WEBGL_WRAPPINGS['REPEAT']

            '''
            tiling = mTex.GetTextureTiling()

            if tiling & MaxPlus.Constants.UMirror:
                wrapS = gltf.WEBGL_WRAPPINGS['MIRRORED_REPEAT']
            elif tiling & MaxPlus.Constants.UWrap:
                wrapS = gltf.WEBGL_WRAPPINGS['REPEAT']

            if tiling & MaxPlus.Constants.VMirror:
                wrapT = gltf.WEBGL_WRAPPINGS['MIRRORED_REPEAT']
            elif tiling & MaxPlus.Constants.VWrap:
                wrapT = gltf.WEBGL_WRAPPINGS['REPEAT']
            '''

            texture['sampler'] = gltf.createSampler(self.output, magFilter, wrapS, wrapT)

            uri = extractTexturePath(mTex)
            texture['source'] = gltf.getImageIndex(self.output, uri)

            #v3dExt = gltf.appendExtension(self.output, 'S8S_v3d_texture_data', texture)

            textures.append(texture)

        if len(textures) > 0:
            self.output['textures'] = textures


    def generateMaterials(self, collector):

        materials = []

        S8S_v3d_data_used = False

        for mMat in collector.materials:
            material = {}

            material['name'] = getName(mMat)
            material['id'] = mMat

            v3dExt = {}
            material['extensions'] = { 'S8S_v3d_material_data' : v3dExt }
            S8S_v3d_data_used = True

            v3dExt['profile'] = 'maya'

            isPbr = isPbrMaterial(mMat)

            if isPbr:
                material['pbrMetallicRoughness'] = {}
                pbr = material['pbrMetallicRoughness']

                matType = cmds.objectType(mMat)

                if matType in ['phong', 'lambert', 'blinn']:

                    trans = list(cmds.getAttr(mMat + '.transparency')[0])
                    transparency = sum(trans) / float(len(trans))

                    tex = extractTexture(mMat, 'color')
                    if tex:
                        pbr['baseColorTexture'] =  {'index': gltf.getTextureIndex(self.output, tex)}
                    else:
                        color = list(cmds.getAttr(mMat + '.color')[0])
                        color.append(1 - transparency)
                        pbr['baseColorFactor'] = color

                    if matType == 'lambert':
                        pbr['metallicFactor'] = 0
                        pbr['roughnessFactor'] = 1
                    elif matType == 'blinn':
                        pbr['metallicFactor'] = cmds.getAttr(mMat + '.specularRollOff')
                        pbr['roughnessFactor'] = cmds.getAttr(mMat + '.eccentricity')
                    elif matType == 'phong':
                        pbr['metallicFactor'] = 1
                        pbr['roughnessFactor'] = 1 - min(1, cmds.getAttr(mMat + '.cosinePower') / 2000)

                elif matType == 'aiStandardSurface' or matType == 'standardSurface':

                    weight = cmds.getAttr(mMat + '.base')

                    tex = extractTexture(mMat, 'baseColor')
                    if tex:
                        pbr['baseColorTexture'] =  {'index': gltf.getTextureIndex(self.output, tex)}
                        pbr['baseColorFactor'] = [weight, weight, weight, 1]
                    else:
                        color = list(cmds.getAttr(mMat + '.baseColor')[0])
                        color = [v * weight for v in color]

                        opacity = list(cmds.getAttr(mMat + '.opacity')[0])
                        opacity = sum(opacity) / float(len(opacity))
                        color.append(opacity)

                        pbr['baseColorFactor'] = color

                    pbr['metallicFactor'] = cmds.getAttr(mMat + '.metalness')
                    pbr['roughnessFactor'] = cmds.getAttr(mMat + '.specularRoughness')

                    normalTex = extractTexture(mMat, 'normalCamera', 'bumpValue')
                    if normalTex:
                        material['normalTexture'] = {'index': gltf.getTextureIndex(self.output, normalTex)}

                        scale = cmds.getAttr(getInputNode(mMat, 'normalCamera') + '.bumpDepth')
                        if scale != 1:
                            material['normalTexture']['scale'] = scale

                    emissiveWeight = cmds.getAttr(mMat + '.emission')

                    emissiveTex = extractTexture(mMat, 'emissionColor')
                    if emissiveTex:
                        material['emissiveTexture'] = {'index': gltf.getTextureIndex(self.output, emissiveTex)}
                        material['emissiveFactor'] = [emissiveWeight, emissiveWeight, emissiveWeight]
                    else:
                        emissive = list(cmds.getAttr(mMat + '.emissionColor')[0])
                        emissive = [v * emissiveWeight for v in emissive]
                        material['emissiveFactor'] = emissive

                elif matType == 'StingrayPBS':
                    tex = extractTexture(mMat, 'TEX_color_map')
                    if tex and cmds.getAttr(mMat + '.use_color_map'):
                        pbr['baseColorTexture'] =  {'index': gltf.getTextureIndex(self.output, tex)}
                    else:
                        color = list(cmds.getAttr(mMat + '.base_color')[0]) + [1]
                        pbr['baseColorFactor'] = color

                    metallicTex = extractTexture(mMat, 'TEX_metallic_map')
                    roughnessTex = extractTexture(mMat, 'TEX_roughness_map')

                    if (metallicTex and cmds.getAttr(mMat + '.use_metallic_map')
                            and roughnessTex and cmds.getAttr(mMat + '.use_roughness_map')):

                        metallicTexPath = extractTexturePath(metallicTex)
                        roghnessTexPath = extractTexturePath(roughnessTex)

                        metalroughTexPath, metalroughQImage = self.createMetallicRoughnessMap(metallicTexPath, roghnessTexPath, getName(mMat))
                        image = self.createImage(metalroughTexPath, metalroughQImage)
                        gltf.appendEntity(self.output, 'images', image)

                        tex = self.createTextureFromImage(image)
                        texIndex = gltf.appendEntity(self.output, 'textures', tex)

                        pbr['metallicRoughnessTexture'] = {'index': texIndex}
                    else:
                        pbr['metallicFactor'] = cmds.getAttr(mMat + '.metallic')
                        pbr['roughnessFactor'] = cmds.getAttr(mMat + '.roughness')

                    normalTex = extractTexture(mMat, 'TEX_normal_map')
                    if normalTex and cmds.getAttr(mMat + '.use_normal_map'):
                        material['normalTexture'] = {'index': gltf.getTextureIndex(self.output, normalTex)}

                    aoTex = extractTexture(mMat, 'TEX_ao_map')
                    if aoTex and cmds.getAttr(mMat + '.use_ao_map'):
                        material['occlusionTexture'] = {'index': gltf.getTextureIndex(self.output, aoTex)}

                    emissiveTex = extractTexture(mMat, 'TEX_emissive_map')

                    if emissiveTex and cmds.getAttr(mMat + '.use_emissive_map'):
                        material['emissiveTexture'] = {'index': gltf.getTextureIndex(self.output, emissiveTex)}
                        emissiveInt = cmds.getAttr(mMat + '.emissive_intensity')
                        material['emissiveFactor'] = [emissiveInt, emissiveInt, emissiveInt]
                    else:
                        emissive = list(cmds.getAttr(mMat + '.emissive')[0])
                        material['emissiveFactor'] = emissive

                material['alphaMode'] = 'OPAQUE'

                if pbr.get('baseColorFactor') and pbr['baseColorFactor'][3] < 1:
                    material['alphaMode'] = 'BLEND'

            materials.append(material)

        if len(collector.skyDomeLights):
            # NOTE: get the first one
            materials.append(self.createWorldMaterial(collector.skyDomeLights[0]))
            S8S_v3d_data_used = True

        if S8S_v3d_data_used:
            gltf.appendExtension(self.output, 'S8S_v3d_material_data')

        if len(materials) > 0:
            self.output['materials'] = materials

    def generateMeshes(self, collector):

        meshes = []

        for mMesh in collector.meshes:

            optimizeAttrs = self.exportSettings['optimizeAttrs']

            shadingGrps = cmds.listConnections(mMesh, type='shadingEngine')
            shadingGrps = list(set(shadingGrps))

            primitives = []

            for group in shadingGrps:
                mMat = cmds.ls(cmds.listConnections(group), materials=True)[0]
                materialIndex = gltf.getMaterialIndex(self.output, mMat)

                meshData = extractMeshData(mMesh, group, optimizeAttrs)
                if not meshData:
                    continue

                indices = meshData['indices']
                positions = meshData['positions']
                normals = meshData['normals']
                uvs = meshData.get('uvs')
                tangents = meshData.get('tangents')

                if len(positions) >= 0xffff:
                    idxCompType = 'UNSIGNED_INT'
                else:
                    idxCompType = 'UNSIGNED_SHORT'

                indicesAccessor = gltf.generateAccessor(self.output, self.binary, indices,
                        idxCompType, len(indices), 'SCALAR', 'ELEMENT_ARRAY_BUFFER')

                vcount = len(positions) // 3

                positionAccessor = gltf.generateAccessor(self.output, self.binary, positions,
                        'FLOAT', vcount, 'VEC3', 'ARRAY_BUFFER')
                normalAccessor = gltf.generateAccessor(self.output, self.binary, normals,
                        'FLOAT', vcount, 'VEC3', 'ARRAY_BUFFER')

                primitive = {
                    'mode': 4,
                    'indices' : indicesAccessor,
                    'attributes' : {
                        'POSITION' : positionAccessor,
                        'NORMAL': normalAccessor
                    },
                    'material' : materialIndex
                }

                if uvs:
                    texcoord0Accessor = gltf.generateAccessor(self.output, self.binary, uvs,
                            'FLOAT', vcount, 'VEC2', 'ARRAY_BUFFER')
                    primitive['attributes']['TEXCOORD_0'] = texcoord0Accessor

                if tangents:
                    tangentAccessor = gltf.generateAccessor(self.output, self.binary, tangents,
                            'FLOAT', vcount, 'VEC4', 'ARRAY_BUFFER')
                    primitive['attributes']['TANGENT'] = tangentAccessor

                primitives.append(primitive)

            mesh = {
                'name': getName(mMesh),
                'id': mMesh,
                'primitives' : primitives
            }

            meshes.append(mesh)

        if len(meshes) > 0:
            self.output['meshes'] = meshes

    def createNode(self, mNode):

        node = {}

        node['name'] = getName(mNode)
        node['id'] = mNode

        node['translation'] = cmds.getAttr(mNode+'.translate')[0]
        node['rotation'] = getRotationQuat(mNode)
        node['scale'] = cmds.getAttr(mNode+'.scale')[0]

        v3dExt = gltf.appendExtension(self.output, 'S8S_v3d_node_data', node)

        childObjs = cmds.listRelatives(mNode, children=True, fullPath=True)
        if childObjs:
            for child in childObjs:
                if isMesh(child):
                    mesh = gltf.getMeshIndex(self.output, child)
                    if mesh >= 0:
                        node['mesh'] = mesh
                elif isCamera(child):
                    camera = gltf.getCameraIndex(self.output, child)
                    if camera >= 0:
                        node['camera'] = camera
                elif isLight(child):
                    light = gltf.getLightIndex(self.output, child)
                    if light >= 0:
                        v3dExt['light'] = light

        v3dExt['hidden'] = not bool(cmds.ls(mNode, visible=True))

        groupNames = mayaUtils.extractGroupNames(mNode)
        if len(groupNames):
            v3dExt['groupNames'] = groupNames

        """
        customProps = extractCustomProps(mNode)
        if customProps:
            node['extras'] = {
                'customProps': customProps
            }
        """

        return node

    def generateNodes(self, collector):
        nodes = []

        for mNode in collector.nodes:

            node = self.createNode(mNode)
            nodes.append(node)

        self.output['nodes'] = nodes

        for mNode in collector.nodes:
            nodeIdx = gltf.getNodeIndex(self.output, mNode)
            if nodeIdx > -1:
                node = self.output['nodes'][nodeIdx]

                children = []

                mChildren = cmds.listRelatives(mNode, children=True, fullPath=True)
                if mChildren:
                    for mChild in mChildren:
                        if mChild in collector.nodes:
                            children.append(gltf.getNodeIndex(self.output, mChild))

                if len(children):
                    node['children'] = children

        if gltf.getLightIndex(self.output, '__DEFAULT__') >= 0:
            # NOTE: find camera node linked to first camera
            for node in nodes:
                if 'camera' in node and node['camera'] == 0:
                    lightNode = self.createDefaultLightNode(node)

                    if 'children' in node:
                        node['children'].append(len(nodes))
                    else:
                        node['children'] = [len(nodes)]

                    nodes.append(lightNode)

    def generateScenes(self, collector):
        scenes = []

        scene = {}
        scene['name'] = 'Scene'
        scene['nodes'] = []

        for mNode in collector.topLevelNodes:
            scene['nodes'].append(gltf.getNodeIndex(self.output, mNode))

        v3dExt = gltf.appendExtension(self.output, 'S8S_v3d_scene_data', scene)

        scene['extras'] = {
            'animFrameRate': getFPS()
        }

        #v3dExt['light'] = gltf.getLightIndex(self.output, '__AMBIENT__')

        settings = self.exportSettings

        worldMatIdx = gltf.getMaterialIndex(self.output, WORLD_NODE_MAT_NAME)
        if worldMatIdx >= 0:
            v3dExt['worldMaterial'] = worldMatIdx

        v3dExt['physicallyCorrectLights'] = True

        v3dExt['useHDR'] = settings['useHDR']
        v3dExt['aaMethod'] = settings['aaMethod']

        v3dExt['shadowMap'] = {
            'type' : 'PCFPOISSON',
            'renderReverseSided' : False,
            'renderSingleSided' : True
        }

        v3dExt['postprocessing'] = []

        if settings['outlineEnabled']:
            effect = {
                'type': 'outline',
                'edgeStrength': settings['edgeStrength'],
                'edgeGlow': settings['edgeGlow'],
                'edgeThickness': settings['edgeThickness'],
                'pulsePeriod': settings['pulsePeriod'],
                'visibleEdgeColor': settings['visibleEdgeColor'] + [1],
                'hiddenEdgeColor': settings['hiddenEdgeColor'] + [1]
            }
            v3dExt['postprocessing'].append(effect)

        scenes.append(scene)

        if len(scenes) > 0:
            self.output['scenes'] = scenes

    def generateScene(self):
        self.output['scene'] = 0

    def createWorldMaterial(self, skyDomeLight):

        envTex = extractTexture(skyDomeLight, 'color')
        intensity = cmds.getAttr(skyDomeLight + '.intensity')

        nodeGraph = { 'nodes': [], 'edges': [] }

        if envTex:
            nodeGraph['nodes'].append({
                'name' : 'World Bitmap',
                'type' : 'AI_SKYDOME_LIGHT',
                'inputs': [intensity],
                'outputs': [],
                'is_active_output': True,
                'texture': gltf.getTextureIndex(self.output, envTex)
            })
        else:
            color = list(cmds.getAttr(skyDomeLight + '.color')[0])
            color = [v * intensity for v in color]
            color += [1]

            nodeGraph['nodes'].append({
                'name' : 'World Bitmap',
                'type' : 'OUTPUT',
                'inputs': [color],
                'outputs': [],
                'is_active_output': True
            })

        worldMat = {
            'name': WORLD_NODE_MAT_NAME,
            'id': WORLD_NODE_MAT_NAME,
            'extensions': {
                'S8S_v3d_material_data': {
                    'profile': 'maya',
                    'nodeGraph': nodeGraph
                }
            }
        }

        return worldMat


    #@profile(immediate=True)
    def run(self, settings):

        self.exportSettings = settings

        if not os.path.exists(settings['filedirectory']):
            os.makedirs(settings['filedirectory'])

        collector = Collector(settings)
        collector.collect()

        self.generateAsset()
        self.generateImages(collector)
        self.generateTextures(collector)
        self.generateMaterials(collector)
        self.generateLights(collector)
        self.generateCameras(collector)
        self.generateMeshes(collector)
        self.generateNodes(collector)
        self.generateAnimations(collector)
        self.generateScenes(collector)
        self.generateScene()
        self.generateBuffers()

        # remove "id" keys
        for key, val in self.output.iteritems():
            if type(val) == list:
                for entity in val:
                    if 'id' in entity:
                        del entity['id']

        indent = None
        separators = separators=(',', ':')

        if (settings['format'] == 'ASCII' and not settings['strip']) or settings['sneakPeek']:
            indent = 4
            separators = separators=(', ', ' : ')

        if settings['format'] == 'ASCII':
            gltfEncoded = unicode(json.dumps(self.output, sort_keys=True, indent=indent,
                                             separators=separators, ensure_ascii=False))

            with open(settings['filepath'], 'w', encoding='utf8') as outfile:
                outfile.write(gltfEncoded)

            if not settings['sneakPeek'] and settings['lzmaEnabled']:
                AppManagerConn.compressLZMA(settings['filepath'])

            if len(self.binary):
                with open(settings['filedirectory'] + settings['binaryfilename'], 'wb') as outfile:
                    outfile.write(self.binary)

                if not settings['sneakPeek'] and settings['lzmaEnabled']:
                    AppManagerConn.compressLZMA(settings['filedirectory'] + settings['binaryfilename'])
        else:

            json_str = json.dumps(self.output, sort_keys=True, indent=indent,
                                  separators=separators, ensure_ascii=False)
            json_bin = bytearray(json_str.encode(encoding='utf8'))
            # 4-byte-aligned
            alignedLen = (len(json_bin) + 3) & ~3
            for i in range(alignedLen - len(json_bin)):
                json_bin.extend(b' ')

            bin_out = bytearray()

            file_length = 12 + 8 + len(json_bin)
            if len(self.binary):
                file_length += 8 + len(self.binary)

            # Magic number
            bin_out.extend(struct.pack('<I', 0x46546C67)) # glTF in binary
            bin_out.extend(struct.pack('<I', 2)) # version number
            bin_out.extend(struct.pack('<I', file_length))
            bin_out.extend(struct.pack('<I', len(json_bin)))
            bin_out.extend(struct.pack('<I', 0x4E4F534A)) # JSON in binary
            bin_out += json_bin

            if len(self.binary):
                bin_out.extend(struct.pack('<I', len(self.binary)))
                bin_out.extend(struct.pack('<I', 0x004E4942)) # BIN in binary
                bin_out += self.binary

            with open(settings['filepath'], 'wb') as outfile:
                outfile.write(bin_out)

            if not settings['sneakPeek'] and settings['lzmaEnabled']:
                AppManagerConn.compressLZMA(settings['filepath'])

        collector.cleanup()

