from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

import imp, os, shutil, tempfile, sys, webbrowser

import maya.cmds as cmds
import maya.mel as mel

import customAttrs, export

from pluginUtils.log import printLog
from pluginUtils.manager import AppManagerConn
from pluginUtils.path import getRoot, getAppManagerHost, getManualURL

join = os.path.join

uiTemplates = set()

def exportGLTF(c):
    # TODO: support startingDirectory param
    filePaths = cmds.fileDialog2(caption='Export glTF', fileFilter='glTF file (*.gltf);;glTF binary file (*.glb)')
    if filePaths:
        exportGLTFPath(filePaths[0])

def exportGLTFPath(filepath=None, sneakPeek=False):
    # NOTE: always reload for dev purposes
    imp.reload(export)

    name, ext = os.path.splitext(os.path.basename(filepath))

    if ext == '.glb':
        gltfFormat = 'BINARY'
    else:
        gltfFormat = 'ASCII'

    filepathBin = name + '.bin'
    filedir = os.path.dirname(filepath) + '/'

    tmpDir = tempfile.mkdtemp(suffix='verge3d')

    exportSettings = {
        'format' : gltfFormat,
        'filepath' : filepath,
        'binaryfilename' : filepathBin,
        'filedirectory': filedir,
        'strip' : True,
        'tmpDir': tmpDir,
        'sneakPeek': sneakPeek,
    }

    customAttrs.parseSettings(exportSettings)

    export.GLTFExporter().run(exportSettings)

def sneakPeek(c):
    prevDir = join(getRoot(), 'player', 'preview')

    # try to cleanup previous
    if os.path.exists(prevDir):
        shutil.rmtree(prevDir, ignore_errors=True)

    if not os.path.exists(prevDir):
        os.mkdir(prevDir)

    exportGLTFPath(join(prevDir, 'sneak_peek.gltf'), True)

    execBrowser(getAppManagerHost() +
            'player/player.html?load=preview/sneak_peek.gltf')

def execBrowser(url):
    # always try to run server before starting browers
    AppManagerConn.start(getRoot(), 'MAYA', False)

    try:
        webbrowser.open(url)
    except BaseException:
        printLog('ERROR', 'Failed to open URL: ' + url)

def runAppManager(c):
    execBrowser(getAppManagerHost())

def runExportSettings(c):
    s = SettingsDialog()
    s.create()

def runUserManual(c):
    cmds.launch(webPage=getManualURL())


def createMenu():
    removeMenu()

    cmds.setParent(mel.eval('$tmp = $gMainWindow'))

    menu = cmds.menu('verge3d', label='Verge3D', tearOff=True)

    cmds.setParent(menu, menu=True)

    cmds.menuItem(label='Export glTF...', command=exportGLTF,
            annotation='Export the scene to glTF 2.0 format')

    cmds.menuItem(divider=True)

    cmds.menuItem(label='Sneek Peak', command=sneakPeek,
            annotation='')

    cmds.menuItem(label='Run App Manager', command=runAppManager,
            annotation='')

    cmds.menuItem(label='Export Settings...', command=runExportSettings,
            annotation='')

    cmds.menuItem(divider=True)

    cmds.menuItem(label='User Manual', command=runUserManual,
            annotation='')

def removeMenu():
    if cmds.control('verge3d', exists=True):
        cmds.deleteUI('verge3d', menu=True)


class SettingsDialog():
    def __init__(self, winName='window'):
        self.winTitle = 'Verge3D Export Settings'
        self.winName = winName

        self.outlineCtls = []

    def outlineChanged(self):
        for ctl in self.outlineCtls:
            enabled = cmds.getAttr('v3dSettings.outlineEnabled')
            cmds.control(ctl, edit=True, enable=enabled)

    def create(self):
        customAttrs.createSettingsNode()

        settingsNode = cmds.ls(type='v3dSettings')

        if not settingsNode:
            return

        if cmds.window(self.winName, exists=True):
            cmds.deleteUI(self.winName)

        cmds.window(self.winName, title=self.winTitle)

        cmds.columnLayout(adjustableColumn=True)

        cmds.frameLayout(label='Common', collapsable=True)

        cmds.attrControlGrp(attribute='v3dSettings.copyright')
        cmds.attrControlGrp(attribute='v3dSettings.lzmaEnabled')
        cmds.attrControlGrp(attribute='v3dSettings.aaMethod')
        cmds.attrControlGrp(attribute='v3dSettings.useHDR')
        cmds.attrControlGrp(attribute='v3dSettings.optimizeAttrs')

        cmds.setParent('..')

        cmds.frameLayout(label='Outlining Effect', collapsable=True)

        cmds.attrControlGrp(attribute='v3dSettings.outlineEnabled', changeCommand=self.outlineChanged)

        self.outlineCtls.append(cmds.attrFieldSliderGrp(attribute='v3dSettings.edgeStrength', enable=False))
        self.outlineCtls.append(cmds.attrFieldSliderGrp(attribute='v3dSettings.edgeGlow', enable=False))
        self.outlineCtls.append(cmds.attrFieldSliderGrp(attribute='v3dSettings.edgeThickness', enable=False))
        self.outlineCtls.append(cmds.attrControlGrp(attribute='v3dSettings.pulsePeriod', enable=False))
        self.outlineCtls.append(cmds.attrColorSliderGrp(attribute='v3dSettings.visibleEdgeColor', enable=False))
        self.outlineCtls.append(cmds.attrColorSliderGrp(attribute='v3dSettings.hiddenEdgeColor', enable=False))

        cmds.setParent('..')

        cmds.frameLayout(label='Animation', collapsable=True)

        cmds.attrControlGrp(attribute='v3dSettings.animExport')
        cmds.attrControlGrp(attribute='v3dSettings.animUsePlaybackRange')
        cmds.attrControlGrp(attribute='v3dSettings.animStartWithZero')

        cmds.showWindow(self.winName)

def animationSettingsUI(nodeAttr, new):

    origParent = cmds.setParent(q=True)

    frameLayoutName = 'AEV3DAnimationSettings'

    if new:
        cmds.frameLayout(frameLayoutName, label='Animation', collapse=False)
    else:
        cmds.setParent(frameLayoutName)

    children = cmds.frameLayout(frameLayoutName, q=True, childArray=True)
    if children:
        for child in children:
            cmds.deleteUI(child)

    ctlRepInf = None
    ctlRepCnt = None
    ctlsCustomRange = []

    def animLoopChanged():
        animLoop = customAttrs.ANIM_LOOP_MAP[cmds.getAttr(nodeAttr+'v3d.animLoop')]
        animRepeatInf = cmds.getAttr(nodeAttr+'v3d.animRepeatInfinite')

        cmds.control(ctlRepInf, edit=True, enable=(animLoop != 'ONCE'))
        cmds.control(ctlRepCnt, edit=True, enable=(animLoop != 'ONCE' and not animRepeatInf))

    def animCustomRangeChanged():
        customRange = cmds.getAttr(nodeAttr+'v3d.animUseCustomRange')
        for ctl in ctlsCustomRange:
            cmds.control(ctl, edit=True, enable=customRange)

    cmds.attrControlGrp(attribute=nodeAttr+'v3d.animAuto')

    cmds.attrControlGrp(attribute=nodeAttr+'v3d.animLoop', changeCommand=animLoopChanged)

    ctlRepInf = cmds.attrControlGrp(attribute=nodeAttr+'v3d.animRepeatInfinite', changeCommand=animLoopChanged)
    ctlRepCnt = cmds.attrControlGrp(attribute=nodeAttr+'v3d.animRepeatCount')
    cmds.attrControlGrp(attribute=nodeAttr+'v3d.animOffset')

    cmds.separator()

    cmds.attrControlGrp(attribute=nodeAttr+'v3d.animUseCustomRange', changeCommand=animCustomRangeChanged)

    ctlsCustomRange.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.animCustomRangeFrom'))
    ctlsCustomRange.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.animCustomRangeTo'))

    cmds.setParent(origParent)

    # update enable status
    animLoopChanged()
    animCustomRangeChanged()

def cameraSettingsUI(nodeAttr, new):

    origParent = cmds.setParent(q=True)

    frameLayoutName = 'AEV3DCameraSettings'

    if new:
        cmds.frameLayout(frameLayoutName, label='Controls', collapse=False)
    else:
        cmds.setParent(frameLayoutName)

    children = cmds.frameLayout(frameLayoutName, q=True, childArray=True)
    if children:
        for child in children:
            cmds.deleteUI(child)

    genericCtls = []
    orbitCtls = []

    def controlsChanged():

        controls = customAttrs.CAMERA_CONTROLS_MAP[cmds.getAttr(nodeAttr+'v3d.controls')]

        for ctl in genericCtls:
            cmds.control(ctl, edit=True, enable=(controls!='NONE'))

        for ctl in orbitCtls:
            cmds.control(ctl, edit=True, enable=(controls=='ORBIT'))

    cmds.attrControlGrp(attribute=nodeAttr+'v3d.controls', changeCommand=controlsChanged)

    genericCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.panningEnabled'))
    genericCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.rotateSpeed'))
    genericCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.moveSpeed'))

    cmds.separator()
    orbitCtls.append(cmds.text(label='Pivot Distance Limits'))
    orbitCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.minDist', label='Min'))
    orbitCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.maxDist', label='Max'))

    cmds.separator()
    orbitCtls.append(cmds.text(label='Vertical Angle Limits'))
    orbitCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.minAngle', label='Min'))
    orbitCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.maxAngle', label='Max'))

    cmds.separator()
    orbitCtls.append(cmds.text(label='Horizontal Angle Limits'))
    orbitCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.minAzimuthAngle', label='Min'))
    orbitCtls.append(cmds.attrControlGrp(attribute=nodeAttr+'v3d.maxAzimuthAngle', label='Max'))

    cmds.setParent(origParent)

    controlsChanged()


def addMelFunctionStubs():
    mel.eval('''
global proc animationSettingsUINew(string $nodeAttr)
{
    python("import interface");
    python("interface.animationSettingsUI('"+$nodeAttr+"', True)");
}
global proc animationSettingsUIReplace(string $nodeAttr)
{
    python("import interface");
    python("interface.animationSettingsUI('"+$nodeAttr+"', False)");
}

global proc cameraSettingsUINew(string $nodeAttr)
{
    python("import interface");
    python("interface.cameraSettingsUI('"+$nodeAttr+"', True)");
}
global proc cameraSettingsUIReplace(string $nodeAttr)
{
    python("import interface");
    python("interface.cameraSettingsUI('"+$nodeAttr+"', False)");
}
''')

def customAttrEditorUI(nodeName):
    type = cmds.objectType(nodeName)

    if type in uiTemplates:
        return

    addMelFunctionStubs()

    if type == 'transform':
        cmds.editorTemplate(beginLayout='Verge3D', collapse=True)
        cmds.editorTemplate('animationSettingsUINew', 'animationSettingsUIReplace', '', callCustom=True)
        cmds.editorTemplate(endLayout=True)

        uiTemplates.add(type)

    elif type == 'camera':
        cmds.editorTemplate(beginLayout='Verge3D', collapse=True)
        cmds.editorTemplate('cameraSettingsUINew', 'cameraSettingsUIReplace', '', callCustom=True)
        cmds.editorTemplate(endLayout=True)

        uiTemplates.add(type)
