
from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

from pluginUtils.log import printLog
import pluginUtils.gltf as gltf

import maya.cmds as cmds
import maya.OpenMaya as OpenMaya


ANIM_ATTR_MAP = {
    'translation': 'translate',
    'rotation': 'rotate',
    'scale': 'scale'
}

ANIM_INTERP_MAP = {
    'spline': 'CUBICSPLINE',
    'linear':'LINEAR',
    'auto': 'LINEAR',
    'fast': 'CUBICSPLINE',
    'slow': 'CUBICSPLINE',
    'step':'STEP',
    'stepnext': 'STEP',
    'fixed':'CUBICSPLINE',
    'clamped': 'CUBICSPLINE',
    'plateau':'CUBICSPLINE'
}

ANIM_TIME_MAP = {
    'game': 15.0,
    'film': 24.0,
    'pal': 25.0,
    'ntsc': 30.0,
    'show': 48.0,
    'palf': 50.0,
    'ntscf': 60.0
}

FLIP_UV = True

def getName(mNode):
    return cmds.ls(mNode, shortNames=True)[0]

def isMesh(mNode):
    return (cmds.objectType(mNode) == 'mesh' and not cmds.getAttr(mNode + '.intermediateObject'))

def isNurbsSurface(mNode):
    return (cmds.objectType(mNode) == 'nurbsSurface' and not cmds.getAttr(mNode + '.intermediateObject'))

def isCamera(mNode):
    return (cmds.objectType(mNode) == 'camera')

def isLight(mNode):
    return (cmds.objectType(mNode) in ('ambientLight', 'directionalLight', 'pointLight', 'spotLight'))

def isSkyDomeLight(mNode):
    return (cmds.objectType(mNode) == 'aiSkyDomeLight')

def isTransformNode(mNode):
    return (cmds.objectType(mNode) == 'transform')

def extractMeshData(mMesh, mShadingGroup, exportTangents):

    cmds.select(mShadingGroup)

    selList = OpenMaya.MSelectionList()
    OpenMaya.MGlobal.getActiveSelectionList(selList)

    if selList.isEmpty():
        cmds.select(deselect=True)
        return None

    mDagPath = OpenMaya.MDagPath()
    mObjComp = OpenMaya.MObject()

    # search for the mesh inside the shader group / set selection
    for i in range(selList.length()):
        selList.getDagPath(i, mDagPath, mObjComp)
        if mDagPath.fullPathName() == mMesh:
            break

    if mObjComp.isNull():
        mMeshIter = OpenMaya.MItMeshPolygon(mDagPath)
    else:
        mMeshIter = OpenMaya.MItMeshPolygon(mDagPath, mObjComp)

    meshFn = OpenMaya.MFnMesh(mDagPath)

    indices = []

    positions = []
    normals = []
    tangents = []
    uvs = []
    colors = []

    mVertIds = OpenMaya.MIntArray()
    mFaceVertIds = OpenMaya.MIntArray()
    mVertPoints = OpenMaya.MPointArray()

    normal = OpenMaya.MVector()

    mNormals = OpenMaya.MFloatVectorArray()
    meshFn.getNormals(mNormals)

    # fixes issue with missing data in UVs
    hasUV = bool(meshFn.numUVs())
    if hasUV:
        uvUtil = OpenMaya.MScriptUtil()
        uvUtil.createFromList([0,0], 2)
        uvPtr = uvUtil.asFloat2Ptr()

        if exportTangents:
            mTangents = OpenMaya.MFloatVectorArray()
            meshFn.getTangents(mTangents)

            mBinormals = OpenMaya.MFloatVectorArray()
            meshFn.getBinormals(mBinormals)

    hasVC = bool(meshFn.numColorSets())
    if hasVC:
        vertexColorList = OpenMaya.MColorArray()
        meshFn.getFaceVertexColors(vertexColorList)


    vNewIdx = 0
    # [] * N does not work!
    vIndMap = [[] for i in range(meshFn.numVertices())]

    while not mMeshIter.isDone():
        mMeshIter.getTriangles(mVertPoints, mVertIds)
        mMeshIter.getVertices(mFaceVertIds)

        faceVertIds = list(mFaceVertIds)

        for vPoint, vIdx in zip(mVertPoints, mVertIds):

            position = (vPoint.x, vPoint.y, vPoint.z)

            faceVertIdx = faceVertIds.index(vIdx)
            normIdx = mMeshIter.normalIndex(faceVertIdx)
            mNorm = mNormals[normIdx]
            normal = (mNorm.x, mNorm.y, mNorm.z)

            if hasUV:
                mMeshIter.getUV(faceVertIdx, uvPtr, meshFn.currentUVSetName())
                u = uvUtil.getFloat2ArrayItem(uvPtr, 0, 0)
                v = uvUtil.getFloat2ArrayItem(uvPtr, 0, 1)

                if FLIP_UV:
                    v = int(v) + (1 - (v % 1))
                uv = (u, v)

                if exportTangents:
                    tangIdx = mMeshIter.tangentIndex(faceVertIdx)

                    mTan = mTangents[tangIdx]
                    mBin = mBinormals[tangIdx]

                    tangent = (mTan.x, mTan.y, mTan.z, 1 if ((mNorm ^ mTan) * mBin > 0) else -1)

            if hasVC:
                color = vertexColorList[vIdx]
                color = (color.r, color.g, color.b)

            indexFound = False

            for i in vIndMap[vIdx]:

                if (positions[i] == position and normals[i] == normal and
                        (not hasUV or uvs[i] == uv) and (not hasVC or colors[i] == color)):
                    indexFound = True
                    indices.append(i)

            if not indexFound:
                indices.append(vNewIdx)
                vIndMap[vIdx].append(vNewIdx)
                vNewIdx += 1

                positions.append(position)
                normals.append(normal)
                if hasUV:
                    uvs.append(uv)

                    if exportTangents:
                        tangents.append(tangent)

                if hasVC:
                    colors.append(color)

        mMeshIter.next()

    meshData = {
        'indices': indices,
        'positions': gltf.flatten(positions),
        'normals': gltf.flatten(normals)
    }

    if hasUV:
        meshData['uvs'] = gltf.flatten(uvs)

        if exportTangents:
            meshData['tangents'] = gltf.flatten(tangents)

    if hasVC:
        meshData['colors'] = gltf.flatten(colors)

    cmds.select(deselect=True)
    return meshData

def getInputNode(mNode, *attrNames):
    currentNode = mNode

    for attr in attrNames:
        colorConn = cmds.listConnections(currentNode + '.' + attr)
        if colorConn and colorConn[0]:
            currentNode = colorConn[0]
        else:
            return None

    return currentNode

def extractTexture(mNode, *attrNames):

    texNode = getInputNode(mNode, *attrNames)

    if texNode and cmds.objectType(texNode) == 'file':
        return texNode
    else:
        return None

def extractTexturePath(mFileNode):
    return cmds.getAttr(mFileNode + '.fileTextureName')

def isPbrMaterial(mMat):
    return True

def getFPS():
    timeUnit = cmds.currentUnit(query=True, time=True)
    fps = ANIM_TIME_MAP.get(timeUnit)
    if fps is None:
        return float(timeUnit.replace('fps', ''))
    else:
        return fps

def getAnimInterpolation(mNode, path, firstKey):
    for axis in ['X','Y','Z']:
        outTangent = cmds.keyTangent(mNode, attribute=ANIM_ATTR_MAP[path]+axis,
                                     time=(firstKey, firstKey), query=True, outTangentType=True)
        if outTangent:
            return ANIM_INTERP_MAP[outTangent[0]]

def getMObject(mNode):
    selectList = OpenMaya.MSelectionList()
    selectList.add(mNode)

    obj = OpenMaya.MObject()
    selectList.getDependNode(0, obj)

    return obj

def getRotationQuat(mNode):

    obj = getMObject(mNode)

    if obj.hasFn(OpenMaya.MFn.kTransform):
        quat = OpenMaya.MQuaternion()
        xform = OpenMaya.MFnTransform(obj)
        xform.getRotation(quat)
        quat.normalizeIt()

    dest = [quat[x] for x in range(4)]
    return dest

def extractAnimData(mNode, path):
    keyframes = cmds.keyframe(mNode, attribute=ANIM_ATTR_MAP[path], query=True, timeChange=True)
    keyframes = sorted(list(set(keyframes)))

    interpolation = getAnimInterpolation(mNode, path, keyframes[0])

    # NOTE: spline interpolation is currently not supported
    if not (interpolation in ['STEP', 'LINEAR']):
        interpolation = 'LINEAR'

    values = []

    if path in ['translation', 'scale']:
        for keyframe in keyframes:
            values.append(cmds.getAttr(mNode+'.'+ANIM_ATTR_MAP[path], time=keyframe)[0])
    else:
        for keyframe in keyframes:
            # NOTE: slow
            cmds.currentTime(keyframe, edit=True)
            values.append(getRotationQuat(mNode))

    fps = getFPS()

    keyframes = [key/fps for key in keyframes]
    values = gltf.flatten(values)

    return (keyframes, values, interpolation)

def extractGroupNames(mNode):

    out = []

    currNode = mNode

    while True:
        parentList = cmds.listRelatives(currNode, parent=True)
        if parentList:
            currNode = parentList[0]
            # only pure transform nodes
            if not cmds.listRelatives(currNode, shapes=True):
                out.append(getName(currNode))
        else:
            break

    return out
