
import imp, mimetypes, os, sys

import maya.OpenMaya as OpenMaya
import maya.OpenMayaMPx as OpenMayaMPx
import maya.cmds as cmds
import maya.mel as mel

TRANSLATOR_NAME = 'Verge3D glTF Exporter'
VENDOR = 'Soft8Soft'
FILE_EXT = 'gltf'

V3D_SETTINGS_NAME = 'v3dSettings'
# from registered ID block 0x00132c40 - 0x00132c7f
V3D_SETTINGS_ID = OpenMaya.MTypeId(0x00132c40)

class V3DSettings(OpenMayaMPx.MPxNode):

    def __init__(self):
        OpenMayaMPx.MPxNode.__init__(self)

def nodeCreator():
    return OpenMayaMPx.asMPxPtr(V3DSettings())

def nodeInitializer():
    import customAttrs
    customAttrs.initSettingsNode(V3D_SETTINGS_NAME)


class V3DTranslator(OpenMayaMPx.MPxFileTranslator):
    def __init__(self):
        OpenMayaMPx.MPxFileTranslator.__init__(self)
    def haveWriteMethod(self):
        return True
    def haveReadMethod(self):
        return False
    def filter(self):
        return '*.{}'.format(FILE_EXT)
    def defaultExtension(self):
        return FILE_EXT
    def writer(self, fileObj, optStr, accessMode):
        from pluginUtils.log import printLog

        fullName = fileObj.fullName()
        try:
            if accessMode == OpenMayaMPx.MPxFileTranslator.kExportAccessMode:
                import interface
                interface.exportGLTFPath(fileObj.fullName())

            elif accessMode == OpenMayaMPx.MPxFileTranslator.kExportActiveAccessMode:
                printLog('ERROR', 'Export selection is not implemented yet')
        except:
            printLog('ERROR', 'Failed to write file information')
            raise

    def reader(self, fileObject, optionString, accessMode):
        raise NotImplementedError()

    def identifyFile(fileObj, buffer, size):
        basename, ext = os.path.splitext(fileObj.fullName())
        if ext not in ['.glb', '.gltf']:
            return OpenMayaMPx.MPxFileTranslator.kNotMyFileType

        return OpenMayaMPx.MPxFileTranslator.kIsMyFileType


def translatorCreator():
    return OpenMayaMPx.asMPxPtr(V3DTranslator())

def devReloadModules():
    import collect, export, customAttrs, interface, mayaUtils

    imp.reload(mayaUtils)
    imp.reload(customAttrs)
    imp.reload(collect)

    imp.reload(interface)
    imp.reload(export)

def onBeforeUnload(*args):
    cmds.flushUndo()

    import customAttrs
    customAttrs.removeSettingsNode()

def initializePlugin(mobject):
    mplugin = OpenMayaMPx.MFnPlugin(mobject, VENDOR, '1.0', 'Any')

    sys.path.append(os.path.normpath(os.path.join(mplugin.loadPath(), 'lib')))
    sys.path.append(os.path.normpath(os.path.join(mplugin.loadPath(), '..', 'python')))

    from pluginUtils.log import printLog
    printLog('INFO', 'Initialize Verge3D plugin')

    mplugin.setName('Verge3D')

    try:
        mplugin.registerFileTranslator(TRANSLATOR_NAME, None, translatorCreator)
    except:
        printLog('ERROR', 'Failed to register translator: {}'.format(TRANSLATOR_NAME))
        raise

    try:
        mplugin.registerNode(V3D_SETTINGS_NAME, V3D_SETTINGS_ID, nodeCreator, nodeInitializer)
    except:
        printLog('ERROR', 'Failed to register node: {}'.format(V3D_SETTINGS_NAME))
        raise

    devReloadModules()

    import interface
    interface.createMenu()

    from pluginUtils.manager import AppManagerConn
    from pluginUtils.path import getRoot

    AppManagerConn.start(getRoot(), 'MAYA', False)
    cmds.scriptJob(event=('quitApplication', AppManagerConn.stop))

    import customAttrs

    customAttrs.assignNodeSettings()
    cmds.scriptJob(event=('SceneOpened', customAttrs.assignNodeSettings))

    OpenMaya.MSceneMessage.addStringArrayCallback(OpenMaya.MSceneMessage.kBeforePluginUnload, onBeforeUnload)

    cmds.callbacks(addCallback=interface.customAttrEditorUI, hook='AETemplateCustomContent', owner=mplugin.name)

def uninitializePlugin(mobject):

    import interface
    interface.removeMenu()

    from pluginUtils.manager import AppManagerConn
    AppManagerConn.stop()

    from pluginUtils.log import printLog

    mplugin = OpenMayaMPx.MFnPlugin(mobject)

    try:
        mplugin.deregisterNode(V3D_SETTINGS_ID)
    except:
        printLog('ERROR', 'Failed to unregister node: {}'.format(V3D_SETTINGS_NAME))
        raise

    try:
        mplugin.deregisterFileTranslator(TRANSLATOR_NAME)
    except:
        printLog('ERROR', 'Failed to unregister translator: {}'.format(TRANSLATOR_NAME))
        raise

