from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

from builtins import (ascii, bytes, chr, dict, filter, hex, input,
                      int, map, next, oct, open, pow, range, round,
                      str, super, zip)

import mimetypes, struct, sys

from .log import printLog

WEBGL_FILTERS = {
    'NEAREST'                : 9728,
    'LINEAR'                 : 9729,
    'NEAREST_MIPMAP_NEAREST' : 9984,
    'LINEAR_MIPMAP_NEAREST'  : 9985,
    'NEAREST_MIPMAP_LINEAR'  : 9986,
    'LINEAR_MIPMAP_LINEAR'   : 9987
};

WEBGL_WRAPPINGS = {
    'CLAMP_TO_EDGE'   : 33071,
    'MIRRORED_REPEAT' : 33648,
    'REPEAT'          : 10497
};

def appendEntity(gltf, name, entity):

    if not gltf.get(name):
        gltf[name] = []

    gltf[name].append(entity)

    # return index
    return (len(gltf[name]) - 1)

def appendExtension(gltf, name, entity=None, extensionData={}, isRequired=False):

    if entity is not None:
        if entity.get('extensions') is None:
            entity['extensions'] = {}

        extensions = entity['extensions']

        if extensions.get(name) is None:
            extensions[name] = {}
        extensions[name].update(extensionData)
        extension = extensions[name]
    else:
        extension = None

    # add to used extensions

    if gltf.get('extensionsUsed') is None:
        gltf['extensionsUsed'] = []

    extensionsUsed = gltf['extensionsUsed']

    if name not in extensionsUsed:
        extensionsUsed.append(name)

    # add to required extensions

    if isRequired:
        if gltf.get('extensionsRequired') is None:
            gltf['extensionsRequired'] = []

        extensionsRequired = gltf['extensionsRequired']

        if name not in extensionsRequired:
            extensionsRequired.append(name)

    return extension

def createSampler(gltf, magFilter, wrapS, wrapT):
    """
    Creates and appends a texture sampler with the given parameters
    """

    if gltf.get('samplers') is None:
        gltf['samplers'] = []

    samplers = gltf['samplers']

    if len(samplers) == 0:
        sampler = {}
        samplers.append(sampler)

    if (magFilter == WEBGL_FILTERS['LINEAR'] and
            wrapS == WEBGL_WRAPPINGS['REPEAT'] and
            wrapT == WEBGL_WRAPPINGS['REPEAT']):
        return 0

    index = 0

    for currentSampler in samplers:
        # pass by empty one
        if currentSampler.get('magFilter') is None or currentSampler.get('wrapS') is None:
            index += 1
            continue

        if (currentSampler['magFilter'] == magFilter and
                currentSampler['wrapS'] == wrapS and
                currentSampler['wrapT'] == wrapT):
            return index

    minFilter = WEBGL_FILTERS['LINEAR_MIPMAP_LINEAR']

    if magFilter == WEBGL_FILTERS['NEAREST']:
        minFilter = WEBGL_FILTERS['NEAREST_MIPMAP_NEAREST']

    sampler = {
        'magFilter' : magFilter,
        'minFilter' : minFilter,
        'wrapS' : wrapS,
        'wrapT' : wrapT
    }

    samplers.append(sampler)

    return len(samplers) - 1

def getSceneIndex(gltf, idname):

    if gltf.get('scenes') is None:
        return -1

    index = 0
    for scene in gltf['scenes']:
        key = 'id' if scene.get('id') != None else 'name'
        if scene.get(key) == idname:
            return index

        index += 1

    return -1

def getNodeIndex(gltf, idname):
    """
    Return the node index in the gltf array.
    """

    if gltf.get('nodes') is None:
        return -1

    index = 0
    for node in gltf['nodes']:
        key = 'id' if node.get('id') != None else 'name'
        if node.get(key) == idname:
            return index

        index += 1

    return -1

def getMeshIndex(gltf, idname):
    """
    Return the mesh index in the gltf array.
    """

    if gltf.get('meshes') is None:
        return -1

    index = 0
    for mesh in gltf['meshes']:
        key = 'id' if mesh.get('id') != None else 'name'
        if mesh.get(key) == idname:
            return index

        index += 1

    return -1


def getMaterialIndex(gltf, idname):
    """
    Return the material index in the gltf array.
    """
    if idname is None:
        return -1

    if gltf.get('materials') is None:
        return -1

    index = 0
    for material in gltf['materials']:
        key = 'id' if material.get('id') != None else 'name'
        if material.get(key) == idname:
            return index

        index += 1

    return -1

def getCameraIndex(gltf, idname):
    """
    Return the camera index in the gltf array.
    """

    if gltf.get('cameras') is None:
        return -1

    index = 0
    for camera in gltf['cameras']:
        key = 'id' if camera.get('id') != None else 'name'
        if camera.get(key) == idname:
            return index

        index += 1

    return -1

def getLightIndex(gltf, idname):
    """
    Return the light index in the gltf array.
    """

    v3dExt = appendExtension(gltf, 'S8S_v3d_data', gltf)

    if v3dExt.get('lights') == None:
        return -1

    lights = v3dExt['lights']

    index = 0
    for light in lights:
        key = 'id' if light.get('id') != None else 'name'
        if light.get(key) == idname:
            return index

        index += 1

    return -1

def getTextureIndex(gltf, idname):

    if gltf.get('textures') is None:
        return -1

    index = 0
    for tex in gltf['textures']:
        key = 'id' if tex.get('id') != None else 'name'
        if tex.get(key) == idname:
            return index

        index += 1

    return -1

def getImageIndex(gltf, idname):

    if gltf.get('images') is None:
        return -1

    index = 0
    for image in gltf['images']:
        key = 'id' if image.get('id') != None else 'name'
        if image.get(key) == idname:
            return index

        index += 1

    return -1

def generateBufferView(gltf, binary, data_buffer, target, alignment):

    if data_buffer is None:
        return -1

    gltf_target_number = [ 34962, 34963 ]
    gltf_target_enums = [ "ARRAY_BUFFER", "ELEMENT_ARRAY_BUFFER" ]

    target_number = 0
    if target in gltf_target_enums:
        target_number = gltf_target_number[gltf_target_enums.index(target)]

    if gltf.get('bufferViews') is None:
        gltf['bufferViews'] = []

    bufferViews = gltf['bufferViews']

    bufferView = {}

    if target_number != 0:
        bufferView['target'] = target_number

    bufferView['byteLength'] = len(data_buffer)

    binary_length = len(binary)

    remainder = 0

    if alignment > 0:
        remainder = binary_length % alignment

    if remainder > 0:
        padding_byte = struct.pack(str('<1b'), 0)
        for i in range(0, alignment - remainder):
            binary.extend(padding_byte)


    bufferView['byteOffset'] = len(binary)
    binary.extend(data_buffer)

    # only have one buffer.
    bufferView['buffer'] = 0

    bufferViews.append(bufferView)

    return len(bufferViews) - 1


def generateAccessor(gltf, binary, data, componentType, count, type, target):

    if data is None:
        printLog('ERROR', 'No data')
        return -1

    gltf_convert_type = [ "b", "B", "h", "H", "I", "f" ]
    gltf_enumNames = [ "BYTE", "UNSIGNED_BYTE", "SHORT", "UNSIGNED_SHORT", "UNSIGNED_INT", "FLOAT" ]
    gltf_convert_type_size = [ 1, 1, 2, 2, 4, 4 ]

    if componentType not in gltf_enumNames:
        printLog('ERROR', 'Invalid componentType ' + componentType)
        return -1

    componentTypeInteger = [ 5120, 5121, 5122, 5123, 5125, 5126 ][gltf_enumNames.index(componentType)]

    convert_type = gltf_convert_type[gltf_enumNames.index(componentType)]
    convert_type_size = gltf_convert_type_size[gltf_enumNames.index(componentType)]

    if count < 1:
        printLog('ERROR', 'Invalid count ' + str(count))
        return -1

    gltf_type_count = [1, 2, 3, 4, 4, 9, 16]
    gltf_type = [ "SCALAR", "VEC2", "VEC3", "VEC4", "MAT2", "MAT3", "MAT4" ]

    if type not in gltf_type:
        printLog('ERROR', 'Invalid type ' + type)
        return -1

    type_count = gltf_type_count[gltf_type.index(type)]


    if gltf.get('accessors') is None:
        gltf['accessors'] = []

    accessors = gltf['accessors']


    accessor = {
        'componentType' : componentTypeInteger,
        'count' : count,
        'type' : type
    }


    minimum = []
    maximum = []

    for component_index in range(0, type_count):
        for component in range(0, count):
            element = data[component * type_count + component_index]

            if component == 0:
                minimum.append(element)
                maximum.append(element)
            else:
                minimum[component_index] = min(minimum[component_index], element)
                maximum[component_index] = max(maximum[component_index], element)

    accessor['min'] = minimum
    accessor['max'] = maximum

    convert_type = '<' + str(count * type_count) + convert_type

    data_buffer = struct.pack(str(convert_type), *data)

    bufferView = generateBufferView(gltf, binary, data_buffer, target, convert_type_size)

    if bufferView < 0:
        printLog('ERROR', 'Invalid buffer view')
        return -1

    accessor['bufferView'] = bufferView

    accessors.append(accessor)

    return len(accessors) - 1

def createAnimChannel(sampler, nodeIndex, path):
    channel = {
        'sampler' : sampler,
        'target': {
            'node': nodeIndex,
            'path': path
        }
    }

    return channel

def createAnimSampler(gltf, binary, keys, values, dim, interpolation='LINEAR'):
    sampler = {}

    sampler['interpolation'] = interpolation

    input = generateAccessor(gltf, binary,
            keys, 'FLOAT', len(keys), 'SCALAR', '')
    sampler['input'] = input

    if dim == 1:
        accessorType = 'SCALAR'
    elif dim == 2:
        accessorType = 'VEC2'
    elif dim == 3:
        accessorType = 'VEC3'
    elif dim == 4:
        accessorType = 'VEC4'

    output = generateAccessor(gltf, binary,
            values, 'FLOAT', len(values) // dim, accessorType, '')
    sampler['output'] = output

    return sampler

def imageMimeType(path):

    # NOTE: add missing HDR mime type to python database
    if sys.version_info < (3, 0):
        mimetypes.add_type('image/vnd.radiance', '.hdr')

    mime = mimetypes.guess_type(path)[0]

    if mime in ['image/jpeg', 'image/bmp', 'image/vnd.radiance', 'image/png']:
        return mime
    else:
        return 'image/png'

def flatten(arr):
    if len(arr) and isinstance(arr[0], tuple):
        return list(sum(arr, ()))
    elif len(arr) and isinstance(arr[0], list):
        return list(sum(arr, []))
    else:
        return []

def getNodeGraph(mat):
    if ('extensions' in mat and 'S8S_v3d_material_data' in mat['extensions']
            and 'nodeGraph' in mat['extensions']['S8S_v3d_material_data']):
        return mat['extensions']['S8S_v3d_material_data']['nodeGraph']
    else:
        return None

