
import { AnimationMixer } from '../animation/AnimationMixer.js';
import { AppUtils } from '../extras/AppUtils.js';
import { BloomPass } from '../postprocessing/BloomPass.js';
import { BrightnessContrastPass } from '../postprocessing/BrightnessContrastPass.js';
import { BokehPass } from '../postprocessing/BokehPass.js';
import { BufferGeometry } from '../core/BufferGeometry.js';
import { Cache } from '../loaders/Cache.js';
import { Clock } from '../core/Clock.js';
import { Compat } from './Compat.js';
import { EffectComposer } from '../postprocessing/EffectComposer.js';
import { Detector } from '../core/Detector.js';
import { DirectionalLightShadow } from '../lights/DirectionalLightShadow.js';
import { FirstPersonControls } from '../controls/FirstPersonControls.js';
import { FlyingControls } from '../controls/FlyingControls.js';
import { FXAAPass } from '../postprocessing/FXAAPass.js';
import { GLTFLoader } from '../loaders/GLTFLoader.js';
import { GrayscalePass } from '../postprocessing/GrayscalePass.js';
import { Line } from '../objects/Line.js';
import { MaterialUtils } from '../materials/MaterialUtils';
import { Mesh } from '../objects/Mesh.js';
import { _Math } from '../math/Math.js';
import { OrbitControls } from '../controls/OrbitControls.js';
import { OutlinePass } from '../postprocessing/OutlinePass.js';
import { Pass } from '../postprocessing/Pass.js';
import { PMREMGenerator } from '../converters/PMREMGenerator.js';
import { PMREMCubeUVPacker } from '../converters/PMREMCubeUVPacker.js';
import { RenderPass } from '../postprocessing/RenderPass.js';
import { RenderUtils } from '../renderers/RenderUtils.js';
import { Scene } from '../scenes/Scene.js';
import { SceneBackground } from '../scenes/SceneBackground.js';
import { SceneUtils } from './SceneUtils.js';
import { ShaderLib } from '../renderers/shaders/ShaderLib.js';
import { ShaderPass } from '../postprocessing/ShaderPass.js';
import { SSAOPass } from '../postprocessing/SSAOPass.js';
import { SSRPass } from '../postprocessing/SSRPass.js';
import { SSAARenderPass } from '../postprocessing/SSAARenderPass.js';
import { ToneMapPass } from '../postprocessing/ToneMapPass.js';
import { Vector2 } from '../math/Vector2.js';
import { Vector3 } from '../math/Vector3.js';
import { WebGLRenderer } from '../renderers/WebGLRenderer.js';
import { WebGLRenderTarget } from '../renderers/WebGLRenderTarget.js';
import { WebGLMultisampleRenderTarget } from '../renderers/WebGLMultisampleRenderTarget.js';

import { VRReticleImage } from '../media.js';
import { isDef } from '../utils.js';

import {
    acceleratedRaycast,
    computeBoundsTree,
    disposeBoundsTree
} from './core/mesh_bvh/index.js';

import {
    EquirectangularReflectionMapping,
    ESMShadowMap,
    PCFSoftShadowMap,
    ViewportFitNone,
    ViewportFitVertical,
    ViewportFitHorizontal,
    ViewportFitAuto,
    RGBAFormat,
    LinearEncoding,
    sRGBEncoding,
    HalfFloatType,
    UnsignedByteType
} from '../constants.js';

var TONE_MAP_PARS = [
    'toneMapping',
    'toneMappingMidTones',
    'toneMappingPhysicalScale',
    // logarithmic max
    'toneMappingBrightness',
    'toneMappingContrast',
    'toneMappingChromaticAdaptation',
    'toneMappingWhiteColor',
    'toneMappingColorDifferentiation',
    'toneMappingExteriorDaylight',
    // physical max
    'toneMappingWhiteBalance',
    'toneMappingHighlights',
    'toneMappingShadows',
    'toneMappingSaturation',
    'toneMappingAperture',
    'toneMappingShutter',
    'toneMappingISO',
    'toneMappingVignetting'
];

var GLTF_PROGRESS_WEIGHT = 0.6;
var COMPILE_PROGRESS_WEIGHT = 0.4;
var ONLOAD_100_TIMEOUT = 16; // ms

var TEX_IMAGE_UNITS_WARN_LIMIT = 8; // maximum for iOS devices

var USE_BVH_RAYCAST = true;

function App(container, ctxSettings, preloader) {

    this.container = container instanceof HTMLElement ? container
            : document.getElementById(container);

    if (!Detector.checkWebGL()) {
        Detector.showWebGLErrorMessage(this.container);
        return;
    }

    this.scene = null;
    this.camera = null;

    this.clock = new Clock();
    this.mixer = null;

    this.renderCallbacks = [];
    this.elapsed = 0;
    // for proper render calls calculation
    this.frame = 0;

    this.preloader = preloader;

    this.worldMaterial = null;
    this.worldCubemapRes = 1024;

    this.xrSession = null;
    this.xrControllers = [];

    // renderer

    ctxSettings = ctxSettings || {};
    if (ctxSettings.alpha === undefined) ctxSettings.alpha = false;
    if (ctxSettings.depth === undefined) ctxSettings.depth = true;
    if (ctxSettings.stencil === undefined) ctxSettings.stencil = true;
    if (ctxSettings.antialias === undefined) ctxSettings.antialias = true;
    if (ctxSettings.premultipliedAlpha === undefined) ctxSettings.premultipliedAlpha = true;
    if (ctxSettings.preserveDrawingBuffer === undefined) ctxSettings.preserveDrawingBuffer = false;
    ctxSettings.xrCompatible = true;

    var canvas = document.createElement('canvas');

    var useWebGL2 = Detector.checkWebGL2();

    // disable buggy WebGL 2.0 implementation on iOS devices
    if (Detector.checkIOS())
        useWebGL2 = false;

    var context = canvas.getContext(useWebGL2 ? 'webgl2' : 'webgl', ctxSettings) ||
            canvas.getContext('experimental-webgl', ctxSettings); // IE11

    this.renderer = new v3d.WebGLRenderer({ canvas: canvas, context: context });
    Compat.prepareRenderer(this.renderer);

    this.clearBkgOnLoad = false;

    this.frameRateDivider = 1;
    this.enableRender = true;
    this.disableRenderTrigger = 0;
    this.ssaaOnPause = false;

    this.renderer.setSize(this.container.offsetWidth, this.container.offsetHeight);

    this.container.appendChild(this.renderer.domElement);

    this.container.classList.add('v3d-container');
    this.renderer.domElement.classList.add('v3d-canvas');

    this.loader = new GLTFLoader();

    if (USE_BVH_RAYCAST) {
        BufferGeometry.prototype.computeBoundsTree = computeBoundsTree;
        BufferGeometry.prototype.disposeBoundsTree = disposeBoundsTree;
        Mesh.prototype.raycast = acceleratedRaycast;
    }

    this.actions = [];
    this.materials = [];

    this._cubeUVEnvRT = null;
    this._resizeCb = null;
    this._loadingTime = 0;

    this._postprocessingSave = null;

    AppUtils.drawWatermark(this);
    AppUtils.addToAppList(this);
}

Object.assign(App.prototype, {

    constructor: App,

    _updateRendererFromGLTF: function(gltf) {

        var floatSupport = Detector.checkHDR(this.renderer);

        if (gltf.renderer.shadowMap) {
            this.renderer.shadowMap.enabled = gltf.renderer.shadowMap.enabled;
            this.renderer.shadowMap.type = gltf.renderer.shadowMap.type;

            if (!floatSupport && this.renderer.shadowMap.type === ESMShadowMap) {
                this.renderer.shadowMap.type = PCFSoftShadowMap;

                gltf.scene.traverse(function(obj) {
                    if (obj.isLight && obj.castShadow && obj.shadow) {
                        // scale bias to make shadows look decent when forcing
                        // them to be PCF instead of ESM
                        obj.shadow.bias /= 100;

                        if (obj.shadow.isDirectionalLightShadowCSM) {
                            // scale radius and initialize non-csm directional shadow
                            obj.shadow.radius *= 100 / obj.shadow.mapSize.x / 2;
                            obj.shadow = new DirectionalLightShadow().copy(obj.shadow);
                            obj.shadow.camera.updateProjectionMatrix();
                        }
                    }
                });
            }
        } else {
            // COMPAT: < 2.3
            this.renderer.shadowMap.enabled = true;
            this.renderer.shadowMap.type = PCFSoftShadowMap;
        }

        // NOTE: does not work good for IBL
        if (gltf.renderer.physicallyCorrectLights)
            this.renderer.physicallyCorrectLights = true;

        this.aaMethod = gltf.renderer.aaMethod || 'AUTO';
        this.useHDR = Boolean(gltf.renderer.useHDR && floatSupport);
        this.renderer.unitsScaleFactor = gltf.renderer.unitsScaleFactor;

        if (gltf.renderer.toneMapping) {
            for (var i = 0; i < TONE_MAP_PARS.length; i++) {
                var par = TONE_MAP_PARS[i];
                if (isDef(gltf.renderer[par]))
                    this.renderer[par] = gltf.renderer[par];
            }
        }
    },

    _updateMeshesRaycastFromGLTF: function(gltf) {
        if (gltf.scene) {
            gltf.scene.traverse(function(obj) {
                var geom = obj.geometry;

                if (obj.isMesh) {
                    // enable optimized raycasting for non-morphed objects only
                    if (geom.computeBoundsTree
                        && Object.keys(geom.morphAttributes).length === 0
                        && !geom.attributes.position.isInterleavedBufferAttribute
                        && (!geom.index || !geom.index.isInterleavedBufferAttribute)) {
                        geom.computeBoundsTree({
                            verbose: false
                        });
                    }
                }
            });
        }
    },

    _updateMaterialsFromGLTF: function(gltf) {

        var scope = this;
        if (gltf.scene) {
            gltf.scene.traverse(function(obj) {
                var mat = obj.material;
                if (mat) {
                    if (scope.useHDR) {
                        mat.useHDR = true;
                        if (mat.isMeshNodeMaterial) {
                            // NOTE: doing it too much
                            mat.updateNodeGraph();
                        }
                    }

                    if (scope.materials.indexOf(mat) == -1) {
                        scope.materials.push(mat);
                    }
                }
            });
        }
    },

    _updateAnimationsFromGLTF: function(gltf, animRoot) {

        var scope = this;

        if (!scope.mixer && gltf.animations && gltf.animations.length) {
            scope.mixer = new AnimationMixer(scope.scene);
        }

        (gltf.animations || []).forEach(function(anim) {
            var animObj = animRoot.getObjectById(anim.nodeId);
            if (animObj) {

                // NOTE: use localRoot for non-root nodes that can be
                // wrongly treated as a root according to the logic in
                // the PropertyBinding.findNode() method
                if (animObj.id != animRoot.id
                        && (animObj.name === "" || animObj.name === "root"
                        || animObj.name === "." || animObj.name === animRoot.name
                        || animObj.name === animRoot.uuid)) {
                    var action = scope.mixer.clipAction(anim.clip, animObj);
                } else {
                    var action = scope.mixer.clipAction(anim.clip, animRoot);
                }

                action.setLoop(anim.loop, anim.repetitions);
                action.startAt(anim.startAt + scope.mixer.time);
                action.clampWhenFinished = true;

                if (anim.auto) {
                    action.play();
                } else {
                    action.stop();
                    action.paused = true;
                }

                scope.actions.push(action);
            }
        });

    },

    _updateSceneCubeUVEnv: function(scene) {
        var scope = this;
        if (scope._cubeUVEnvRT) {
            scope._traverseSceneForEnvUpdates(scene, function(obj) {
                obj.material.envMap = scope._cubeUVEnvRT.texture;
                obj.material.needsUpdate = true;
            });
        }
    },

    _traverseSceneForEnvUpdates: function(scene, cb) {
        scene.traverse(function(obj) {
            if (obj.isMesh && obj.material && (obj.material.isMeshStandardMaterial
                    || obj.material.isMeshPhysicalMaterial
                    || (obj.material.isMeshNodeMaterial
                    && 'PHYSICAL' in obj.material.defines))) {
                cb(obj);
            }
        });
    },

    _traverseSceneForTexUniforms: function(scene, cb) {
        scene.traverse(function(obj) {
            if (obj.material) {
                var objMats = Array.isArray(obj.material) ? obj.material : [obj.material];
                objMats.forEach(function(mat) {
                    if (mat.program !== undefined) {
                        cb(obj, mat, mat.program.getTexUniformCount());
                    }
                });
            }
        });
    },

    _checkMSAA: function(renderer, requestedSamples) {
        return Math.min(requestedSamples, renderer.capabilities.maxSamples);
    },

    _getActiveSSAAPass: function(composer) {
        if (composer.passes[0] instanceof SSAARenderPass)
            return composer.passes[0];
        else
            return null;
    },

    _precompileSceneAsync: function(scene, camera, renderTarget, progressCb) {

        var scope = this;

        var currentRenderTarget = scope.renderer.getRenderTarget();
        scope.renderer.setRenderTarget(renderTarget);

        scope.renderer.compileAsync(scene, camera, function(percentage) {

            if (percentage >= 1) {
                scope.renderer.setRenderTarget(currentRenderTarget);
            }

            if (progressCb) progressCb(percentage * 100);

        });

    },

    load: function(url, loadOkCb, loadErrorCb, autoStart) {

        console.warn('v3d.App.load has been deprecated. Use v3d.App.loadScene instead.');

        var scope = this;

        scope.loadScene(url, function(loadedScene) {
            loadOkCb(loadedScene);
            if (autoStart) scope.run();
        }, null, loadErrorCb);

    },

    loadScene: function(url, loadCb, progressCb, errorCb) {

        var scope = this;

        if (!scope.renderer) {
            if (errorCb) errorCb('WebGL not found');
            return;
        }

        scope._loadingTime = performance.now();

        this.loader.load(url, function(gltf) {

            // setup scene/camera
            scope.scene = gltf.scene || new Scene();

            var sceneBox = null;

            if (gltf.cameras && gltf.cameras.length) {
                scope.camera = gltf.cameras[0];
            } else {
                sceneBox = SceneUtils.calcSceneBox(scope.scene);

                scope.camera = SceneUtils.createDefaultCamera(sceneBox,
                        scope.container.offsetWidth / scope.container.offsetHeight);
                scope.scene.add(scope.camera);
            }

            // not a Verge3D asset
            if (!scope.camera.controls) {
                // for optimization
                sceneBox = sceneBox || SceneUtils.calcSceneBox(scope.scene);

                SceneUtils.assignDefaultControls(scope.camera, SceneUtils.calcSceneBox(scope.scene));
                scope.camera.viewportFit.type = ViewportFitVertical;
            }

            scope._updateRendererFromGLTF(gltf);
            scope._updateMaterialsFromGLTF(gltf);
            scope._updateAnimationsFromGLTF(gltf, scope.scene);

            if (USE_BVH_RAYCAST) {
                scope._updateMeshesRaycastFromGLTF(gltf);
            }

            var wMat = gltf.world.material;
            if (wMat) {
                scope.updateEnvironment(wMat);
                scope.worldMaterial = wMat;
            }

            scope.initPostprocessing();

            if (scope.onResize) scope.onResize();

            // make matrices up to date for possible use in the load callback
            scope.scene.updateMatrixWorld();

            var rt = scope.postprocessing ? scope.postprocessing.composer.renderTarget1 : null;
            scope._precompileSceneAsync(scope.scene, scope.camera, rt,
                function(percentage) {

                    var totalPercentage = (GLTF_PROGRESS_WEIGHT * 100 + COMPILE_PROGRESS_WEIGHT * percentage)
                            / (GLTF_PROGRESS_WEIGHT + COMPILE_PROGRESS_WEIGHT);
                    if (progressCb) progressCb(totalPercentage);
                    if (scope.preloader) scope.preloader.onUpdate(totalPercentage);

                    if (totalPercentage >= 100) {

                        scope._traverseSceneForTexUniforms(scope.scene, function(obj, mat, count) {
                            if (count > TEX_IMAGE_UNITS_WARN_LIMIT) {
                                console.warn(`v3d.App: Material "${mat.name}" on object `
                                        + `"${obj.name}" exceeds iOS limit of `
                                        + `${TEX_IMAGE_UNITS_WARN_LIMIT} textures `
                                        + `(has ${count}).`);
                            }
                        });

                        if (loadCb) {
                            setTimeout(function() {
                                scope._loadingTime = performance.now() - scope._loadingTime;
                                loadCb(gltf.scene);
                            }, ONLOAD_100_TIMEOUT);
                        }

                    }

                }
            );

            if (scope.clearBkgOnLoad) scope.scene.background = null;

        }, function(percentage) {

            var totalPercentage = GLTF_PROGRESS_WEIGHT * percentage
                    / (GLTF_PROGRESS_WEIGHT + COMPILE_PROGRESS_WEIGHT);
            if (progressCb) progressCb(totalPercentage);
            if (scope.preloader) scope.preloader.onUpdate(totalPercentage);

        }, function(error) {

            console.error(error);
            if (errorCb) errorCb(error);

        });

        // remove previous resize callback if exists
        window.removeEventListener('resize', this._resizeCb, false);
        this._resizeCb = function() { if (scope.onResize) scope.onResize(); }
        window.addEventListener('resize', this._resizeCb, false);
    },

    appendScene: function(url, loadCb, progressCb, errorCb, loadCameras, loadLights) {

        var scope = this;

        if (!scope.renderer) {
            if (errorCb) errorCb('WebGL not found');
            return;
        }

        if (loadCameras === undefined) loadCameras = true;
        if (loadLights === undefined) loadLights = true;

        this.loader.load(url, function(gltf) {

            var loadedScene = gltf.scene || new Scene();

            var objsToUnload = [];
            loadedScene.traverse(function(obj) {
                if (obj.isCamera && !loadCameras || obj.isLight && !loadLights) {
                    objsToUnload.push(obj);
                }
            });

            objsToUnload.forEach(function(obj) {
                scope.unload(obj);
            });

            if (scope.scene) {
                scope.scene.add(loadedScene);

                scope._updateMaterialsFromGLTF(gltf);
                scope._updateAnimationsFromGLTF(gltf, loadedScene);
                if (USE_BVH_RAYCAST) {
                    scope._updateMeshesRaycastFromGLTF(gltf);
                }

                scope._updateSceneCubeUVEnv(loadedScene);

                // make matrices up to date for possible use in the load callback
                scope.scene.updateMatrixWorld();
            }

            var rt = scope.postprocessing ? scope.postprocessing.composer.renderTarget1 : null;
            // use the whole scene because it can be affected by something from
            // the dynamically loaded scene, e.g. lights
            var compiledScene = scope.scene || loadedScene;
            scope._precompileSceneAsync(compiledScene, scope.camera, rt,
                function(percentage) {

                    var totalPercentage = (GLTF_PROGRESS_WEIGHT * 100 + COMPILE_PROGRESS_WEIGHT * percentage)
                            / (GLTF_PROGRESS_WEIGHT + COMPILE_PROGRESS_WEIGHT);
                    if (progressCb) progressCb(totalPercentage);
                    if (scope.preloader) scope.preloader.onUpdate(totalPercentage);

                    if (totalPercentage >= 100) {

                        scope._traverseSceneForTexUniforms(loadedScene, function(obj, mat, count) {
                            if (count > TEX_IMAGE_UNITS_WARN_LIMIT) {
                                console.warn(`v3d.App: Material "${mat.name}" on object `
                                        + `"${obj.name}" exceeds iOS limit of `
                                        + `${TEX_IMAGE_UNITS_WARN_LIMIT} textures `
                                        + `(has ${count}).`);
                            }
                        });

                        if (loadCb) {
                            setTimeout(function() {
                                loadCb(loadedScene);
                            }, ONLOAD_100_TIMEOUT);
                        }

                    }

                }
            );

        }, function(percentage) {

            var totalPercentage = GLTF_PROGRESS_WEIGHT * percentage
                    / (GLTF_PROGRESS_WEIGHT + COMPILE_PROGRESS_WEIGHT);
            if (progressCb) progressCb(totalPercentage);
            if (scope.preloader) scope.preloader.onUpdate(totalPercentage);

        }, function(error) {

            console.error(error);
            if (errorCb) errorCb(error);

        });
    },

    unload: function(rootObj) {

        var scope = this;

        if (!rootObj) rootObj = scope.scene;

        function disposeObjResources(obj) {
            if (obj.isAnnotation) {
                obj.dispose();
            } else if (obj.isLight && obj.shadow && obj.shadow.isLightShadow) {
                obj.shadow.dispose();
            } else if (obj.isCamera && obj.controls && obj.controls.orbitTarget) {
                disposeObjResources(obj.controls.orbitTarget);
            } else if (obj.isMesh) {
                obj.geometry.dispose();

                if (obj.geometry.disposeBoundsTree) {
                    obj.geometry.disposeBoundsTree();
                }

                var mat = obj.material;
                MaterialUtils.disposeTextures(mat);
                mat.dispose();
            }
        }

        if (rootObj === scope.scene) {
            // full scene cleanup

            if (scope.scene) {
                scope.scene.traverse(disposeObjResources);
                scope.disposeEnvironment();
            }
            scope.scene = null;

            if (scope.controls && scope.controls.dispose) {
                scope.controls.dispose();
            }
            scope.controls = null;
            scope.camera = null;

            scope.materials = [];

            if (scope.mixer) {
                scope.mixer.stopAllAction();
                scope.actions.forEach(function(action) {
                    scope.mixer.uncacheAction(action.getClip(), action.getRoot());
                });
            }
            scope.mixer = null;
            scope.actions = [];

            if (scope.postprocessing) {
                if (scope.postprocessing.composer) {
                    for (var name in scope.postprocessing) {
                        var ppEffect = scope.postprocessing[name];
                        if (ppEffect instanceof Pass || ppEffect instanceof EffectComposer) {
                            ppEffect.dispose();
                        }
                    }
                }
                scope.disablePostprocessing();
            }
            scope.postprocessing = null;

            if (scope.renderer) scope.renderer.disposeInternalCaches();

        } else {

            if (scope.scene) {
                rootObj.traverse(disposeObjResources);
                if (rootObj.parent) rootObj.parent.remove(rootObj);
            }

            scope.materials = [];
            if (scope.scene) {
                scope.scene.traverse(function(obj) {
                    var mat = obj.material;
                    if (mat && scope.materials.indexOf(mat) == -1) {
                        scope.materials.push(mat);
                    }
                });
            }

            if (scope.mixer) {
                for (var i = scope.actions.length - 1; i >= 0; i--) {
                    var action = scope.actions[i];
                    if (!SceneUtils.checkActionIsUsed(scope.scene, action)) {
                        action.stop();
                        scope.mixer.uncacheAction(action.getClip(), action.getRoot());
                        scope.actions.splice(i, 1);
                    }
                }
            }

        }

    },

    dispose: function() {
        if (this.scene) this.unload();

        if (this.renderer) {
            this.renderer.forceContextLoss();
            this.renderer.dispose();
        }
        this.renderer = null;

        if (this.container) {
            while (this.container.firstChild) {
                this.container.removeChild(this.container.firstChild);
            }
        }

        window.removeEventListener('resize', this._resizeCb, false);

        AppUtils.removeFromAppList(this);

        Cache.clear();
    },

    getWidth: function() {
        return this.container.offsetWidth;
    },

    getHeight: function() {
        return this.container.offsetHeight;
    },

    onResize: function() {

        function recalcVerticalFov(oldFov, oldAspect, newAspect) {
            return _Math.RAD2DEG * 2 * Math.atan(
                    Math.tan(_Math.DEG2RAD * oldFov / 2) * oldAspect / newAspect);
        }

        return function() {

            var width = this.getWidth();
            var height = this.getHeight();

            this.renderer.setSize(width, height);

            if (this.postprocessing) {
                var pixRatio = this.renderer.getPixelRatio();
                this.postprocessing.composer.setSize(width * pixRatio, height * pixRatio);
            }

            if (this.scene) {

                var aspect = width / height;
                var cam = this.camera;

                if (cam.isPerspectiveCamera) {

                    switch (cam.viewportFit.type) {
                    case ViewportFitVertical:
                        cam.aspect = aspect;
                        break;
                    case ViewportFitHorizontal:
                        cam.fov = recalcVerticalFov(cam.fov,
                                cam.aspect, aspect);
                        cam.aspect = aspect;
                        break;
                    case ViewportFitAuto:

                        var oldAspectIsLess = cam.aspect < cam.viewportFit.initialAspect;
                        var newAspectIsLess = aspect < cam.viewportFit.initialAspect;

                        if (oldAspectIsLess && newAspectIsLess)
                            cam.fov = recalcVerticalFov(cam.fov, cam.aspect, aspect);
                        else if (oldAspectIsLess && !newAspectIsLess)
                            cam.fov = recalcVerticalFov(cam.fov, cam.aspect, cam.viewportFit.initialAspect);
                        else if (!oldAspectIsLess && newAspectIsLess)
                            cam.fov = recalcVerticalFov(cam.fov, cam.viewportFit.initialAspect, aspect);

                        cam.aspect = aspect;
                        break;
                    case ViewportFitNone:
                    default:
                        break;
                    }

                } else if (cam.isOrthographicCamera) {

                    switch (cam.viewportFit.type) {
                    case ViewportFitVertical:
                        var horizSize = cam.top * aspect;
                        cam.left = -horizSize;
                        cam.right = horizSize;
                        break;
                    case ViewportFitHorizontal:
                        var vertSize = cam.right / aspect;
                        cam.bottom = -vertSize;
                        cam.top = vertSize;
                        break;
                    case ViewportFitAuto:

                        var oldAspectIsLess = (cam.right - cam.left) /
                                (cam.top - cam.bottom) < cam.viewportFit.initialAspect;
                        var newAspectIsLess = aspect < cam.viewportFit.initialAspect;

                        if (oldAspectIsLess && newAspectIsLess)
                            var horizSize = cam.right;
                        else if (oldAspectIsLess && !newAspectIsLess)
                            var horizSize = cam.right * aspect / cam.viewportFit.initialAspect;
                        else if (!oldAspectIsLess && newAspectIsLess)
                            var horizSize = cam.top * cam.viewportFit.initialAspect;
                        else
                            var horizSize = cam.top * aspect;

                        cam.left = -horizSize;
                        cam.right = horizSize;
                        cam.bottom = -horizSize / aspect;
                        cam.top = horizSize / aspect;

                        break;
                    case ViewportFitNone:
                    default:
                        break;
                    }

                }

                cam.updateProjectionMatrix();

            }

        }
    }(),

    run: function() {

        if (this.preloader) {
            this.preloader.onFinish();
            this.preloader = null;
        }
        this.animate();

    },

    animate: function() {
        var scope = this;

        var cb = function() {

            var elapsed = scope.clock.getDelta();
            scope.elapsed = elapsed;

            if (scope.mixer)
                scope.mixer.update(elapsed);

            if (scope.controls && !scope.xrSession)
                scope.controls.update(elapsed);

            // handle XR controllers
            for (var i = 0; i < scope.xrControllers.length; i++) {
                var controller = scope.xrControllers[i];

                // disable ray line on gaze controls
                for (var j = 0; j < controller.children.length; j++) {
                    var child = controller.children[j];

                    // handle only specially created objects
                    if (child.name != 'XR_CONTROLLER_RAY' && child.name != 'XR_CONTROLLER_RETICLE')
                        continue;

                    if (controller.userData.inputSource &&
                            controller.userData.inputSource.targetRayMode == 'gaze') {
                        child.visible = (child.name == 'XR_CONTROLLER_RAY') ? false : true;
                    } else {
                        child.visible = (child.name == 'XR_CONTROLLER_RAY') ? true : false;
                    }
                }
            }


            var rbs = scope.renderCallbacks;
            for (var i = 0; i < rbs.length; i++)
                rbs[i](elapsed, scope.clock.elapsedTime);

            if ((scope.frame % scope.frameRateDivider) == 0 && scope.enableRender) {
                scope.render();
            }

            scope.frame++;

            if (scope.disableRenderTrigger && --scope.disableRenderTrigger == 0)
                scope.enableRender = false;
        }

        scope.renderer.setAnimationLoop(cb);
    },

    enableRendering: function() {
        this.disableRenderTrigger = 0;
        this.enableRender = true;
        if (this.ssaaOnPause)
            this.enableSSAA(0, true);
    },

    disableRendering: function(after) {
        after = after || 0;

        if ((after || this.ssaaOnPause) && this.enableRender && this.disableRenderTrigger == 0) {
            this.disableRenderTrigger = this.ssaaOnPause ? 16 : after;
            if (this.ssaaOnPause)
                this.enableSSAA(4, true);
        } else if (after == 0) {
            this.enableRender = false;
        }
    },

    setFrameRateDivider: function(divider) {
        this.frameRateDivider = divider;
    },

    render: function() {
        if (this.postprocessing) {
            var composer = this.postprocessing.composer;
            var ssaaPass = this._getActiveSSAAPass(composer);

            if (ssaaPass && ssaaPass.iterative) {
                if (ssaaPass.isLastIterativeFrame())
                    composer.renderToScreen = true;
                else
                    composer.renderToScreen = false;
            } else {
                composer.renderToScreen = true;
            }


            composer.render(this.elapsed);
        } else if (this.scene && this.camera) {
            this.renderer.render(this.scene, this.camera);
        }
    },

    enableControls: function(element) {
        var camera = this.camera;

        if (!camera.controls)
            return;

        switch (camera.controls.type) {
        case 'ORBIT':
            this.controls = new OrbitControls(camera, element || this.renderer.domElement);
            this.controls.targetObj = camera.controls.orbitTarget;
            this.controls.minDistance = camera.controls.orbitMinDistance;
            this.controls.maxDistance = camera.controls.orbitMaxDistance;
            this.controls.minPolarAngle = camera.controls.orbitMinPolarAngle;
            this.controls.maxPolarAngle = camera.controls.orbitMaxPolarAngle;
            this.controls.minAzimuthAngle = camera.controls.orbitMinAzimuthAngle;
            this.controls.maxAzimuthAngle = camera.controls.orbitMaxAzimuthAngle;
            break;
        case 'FLYING':
            this.controls = new FlyingControls(camera, this.renderer.domElement);

            this.controls.panSpeedTouch *= camera.controls.moveSpeed;
            this.controls.zoomSpeedKey *= camera.controls.moveSpeed;
            break;
        case 'FIRST_PERSON':
            this.controls = new FirstPersonControls(camera, this.renderer.domElement);
            this.controls.collisionMeshes = [];

            var scope = this;

            this.scene.traverse(function(obj) {
                var mat = obj.material;
                if (mat && camera.controls.collisionMaterial && mat.name == camera.controls.collisionMaterial.name) {
                    scope.controls.collisionMeshes.push(obj);
                }
            });

            scope.controls.gazeLevel = camera.controls.gazeLevel;
            scope.controls.storyHeight = camera.controls.storyHeight;

            this.controls.zoomSpeedKey *= camera.controls.moveSpeed;

            break;
        default:
            this.controls = null;
            break;
        }

        if (this.controls) {
            this.controls.enablePan = camera.controls.enablePan;

            this.controls.rotateSpeed *= camera.controls.rotateSpeed;
            this.controls.rotateSpeedTouch *= camera.controls.rotateSpeed;

            this.controls.panSpeed *= camera.controls.moveSpeed;
            this.controls.panSpeedKey *= camera.controls.moveSpeed;

            this.controls.zoomSpeed *= camera.controls.moveSpeed;
            this.controls.zoomSpeedTouch *= camera.controls.moveSpeed;
        }
    },

    setCamera: function(camera) {
        this.camera = camera;

        if (this.postprocessing) {
            var passes = this.postprocessing.composer.passes;
            for (var i = 0; i < passes.length; i++)
                passes[i].setCamera(camera);
        }

        if (this.controls)
            this.controls.dispose();

        this.enableControls();
        this.onResize();
    },

    getCamera: function() {
        return this.camera;
    },

    cleanup: function() {

        console.warn('v3d.App.cleanup has been deprecated. Use v3d.App.unload or v3d.App.dispose instead.');

        if (this.container) {
            while (this.container.firstChild) {
                this.container.removeChild(this.container.firstChild);
            }
        }

        if (!this.loader || !this.mixer)
            return;

        this.mixer.stopAllAction();
    },

    initPostprocessing: function() {
        // also enable ToneMap/FXAA for HDR pipeline
        if (this.scene.postprocessing && this.scene.postprocessing.length || this.useHDR || this.aaMethod != 'AUTO')
            this.enablePostprocessing(this.scene.postprocessing || []);
    },

    enablePostprocessing: function(effects) {

        var renderTargetParams = {
            format: RGBAFormat,
            stencilBuffer: false
        }

        if (this.useHDR) {
            renderTargetParams.type = HalfFloatType;
            renderTargetParams.encoding = LinearEncoding;
        } else {
            renderTargetParams.type = UnsignedByteType;
            renderTargetParams.encoding = sRGBEncoding;
        }

        if (!this.postprocessing) {
            this.postprocessing = {};
            var renderPass = new RenderPass(this.scene, this.camera);
            this.postprocessing.renderPass = renderPass;

            var aaSamples = 0;
            switch (this.aaMethod) {
            case 'AUTO':
            case 'MSAA4':
                aaSamples = this._checkMSAA(this.renderer, 4);
                break;
            case 'MSAA8':
                aaSamples = this._checkMSAA(this.renderer, 8);
                break;
            case 'MSAA16':
                aaSamples = this._checkMSAA(this.renderer, 16);
                break;
            }

            if (Detector.checkSwiftShader(this.renderer)) {
                aaSamples = 0;
                console.warn('v3d.App: disabling buggy multisampling on SwiftShader renderer');
            }

            if (aaSamples) {
                var renderTarget = new WebGLMultisampleRenderTarget(this.getWidth(), this.getHeight(),
                        renderTargetParams);
                renderTarget.samples = aaSamples;
            } else {
                var renderTarget = new WebGLRenderTarget(this.getWidth(), this.getHeight(),
                        renderTargetParams);
            }

            renderTarget.texture.name = 'EffectComposer.rt1';

            var composer = new EffectComposer(this.renderer, renderTarget);
            composer.addPass(renderPass);

            this.postprocessing.composer = composer;

            if (this.useHDR) {
                var toneMapPass = new ToneMapPass();
                composer.addPass(toneMapPass);
                this.postprocessing.toneMapPass = toneMapPass;
            }

            if (!aaSamples) {
                var fxaaPass = new FXAAPass(this.scene, this.camera);
                composer.addPass(fxaaPass);
                this.postprocessing.fxaaPass = fxaaPass;
            }

        } else {
            var composer = this.postprocessing.composer;
        }

        // cleanup unnecessary copy pass
        if (effects.length && this.postprocessing.copyPass) {
            composer.passes.splice(composer.passes.indexOf(this.postprocessing.copyPass), 1);
            this.postprocessing.copyPass = null;
        }

        for (var i = 0; i < effects.length; i++) {
            var effect = effects[i];

            var insertIndex = composer.passes.length - 1;

            // no FXAA
            if (!(composer.passes[insertIndex] instanceof FXAAPass))
                insertIndex++

            // before Tonemap pass
            if (this.useHDR)
                insertIndex--;

            switch (effect.type) {
            case 'bloom':
                var bloomStrength = effect.strength;
                var bloomRadius = effect.radius;
                var bloomThreshold = effect.threshold;

                if (!this.postprocessing.bloomPass) {
                    var bloomPass = new BloomPass(new Vector2(this.getWidth(), this.getHeight()),
                            bloomStrength, bloomRadius, bloomThreshold, renderTargetParams);

                    composer.insertPass(bloomPass, insertIndex);
                    this.postprocessing.bloomPass = bloomPass;
                } else {
                    var bloomPass = this.postprocessing.bloomPass;

                    bloomPass.strength = bloomStrength;
                    bloomPass.radius = bloomRadius;
                    bloomPass.threshold = bloomThreshold;
                }

                break;

            case 'brightnessContrast':
                if (!this.postprocessing.brightnessContrastPass) {
                    var brightnessContrastPass = new BrightnessContrastPass();

                    composer.insertPass(brightnessContrastPass, insertIndex);
                    this.postprocessing.brightnessContrastPass = brightnessContrastPass;
                } else
                    var brightnessContrastPass = this.postprocessing.brightnessContrastPass;

                brightnessContrastPass.brightness = effect.brightness;
                brightnessContrastPass.contrast = effect.contrast;

                break;

            case 'dof':
                if (!this.postprocessing.bokehPass) {
                    var bokehPass = new BokehPass(this.scene, this.camera, {
                        width: this.getWidth(),
                        height: this.getHeight()
                    });
                    composer.insertPass(bokehPass, insertIndex);
                    this.postprocessing.bokehPass = bokehPass;
                } else {
                    var bokehPass = this.postprocessing.bokehPass;
                }

                bokehPass.focus = effect.focus;
                bokehPass.aperture = effect.aperture;
                bokehPass.maxblur = effect.maxblur;
                bokehPass.depthLeakThreshold = effect.depthLeakThreshold;

                break;

            case 'grayscale':
                if (!this.postprocessing.grayscalePass) {
                    var grayscalePass = new GrayscalePass();

                    composer.insertPass(grayscalePass, insertIndex);
                    this.postprocessing.grayscalePass = grayscalePass;
                }

                break;

            case 'outline':
                if (!this.postprocessing.outlinePass) {
                    var outlinePass = new OutlinePass(
                            new Vector2(this.getWidth(), this.getHeight()),
                            this.scene, this.camera);
                    composer.insertPass(outlinePass, insertIndex);
                    this.postprocessing.outlinePass = outlinePass;
                } else
                    outlinePass = this.postprocessing.outlinePass;

                outlinePass.edgeStrength = effect.edgeStrength;
                outlinePass.edgeGlow = effect.edgeGlow;
                outlinePass.edgeThickness = effect.edgeThickness;
                outlinePass.pulsePeriod = effect.pulsePeriod;
                outlinePass.visibleEdgeColor.fromArray(effect.visibleEdgeColor.slice(0, 3))
                outlinePass.hiddenEdgeColor.fromArray(effect.hiddenEdgeColor.slice(0, 3));

                break;

            case 'ssao':
                if (!Detector.checkDepthTex(this.renderer)) {
                    console.warn('v3d.App: disabling SSAO since your hardware does not support depth textures');
                    break;
                }

                if (!this.postprocessing.ssaoPass) {
                    var ssaoPass = new SSAOPass(this.scene, this.camera, this.getWidth(), this.getHeight());
                    composer.insertPass(ssaoPass, insertIndex);
                    this.postprocessing.ssaoPass = ssaoPass;
                } else
                    var ssaoPass = this.postprocessing.ssaoPass;

                ssaoPass.kernelRadius = effect.radius;
                ssaoPass.minDistance = effect.minDistance || 0.005;
                ssaoPass.maxDistance = effect.maxDistance || 0.1;

                break;
            case 'ssr':
                if (!Detector.checkHalfFloatTex(this.renderer, false)) {
                    console.warn('v3d.App: disabling SSR since your hardware does not support float textures');
                    break;
                }

                if (isDef(effect.useRefract))
                    var useRefract = effect.useRefract;
                else
                    var useRefract = false;

                if (!(useRefract ? this.postprocessing.ssrPassRefract : this.postprocessing.ssrPassReflect)) {
                    var ssrPass = new SSRPass(this.scene, this.camera);
                    composer.insertPassAfter(ssrPass, [
                            this.postprocessing.renderPass,
                            this.postprocessing.ssrPassRefract,
                            this.postprocessing.ssrPassReflect]);
                    if (useRefract)
                        this.postprocessing.ssrPassRefract = ssrPass;
                    else
                        this.postprocessing.ssrPassReflect = ssrPass;
                } else {
                    var ssrPass = useRefract ? this.postprocessing.ssrPassRefract :
                            this.postprocessing.ssrPassReflect;
                }

                ssrPass.useRefract = useRefract;

                if (isDef(effect.objects))
                    ssrPass.objects = effect.objects;
                if (isDef(effect.intensity))
                    ssrPass.intensity = effect.intensity;
                if (isDef(effect.steps))
                    ssrPass.steps = effect.steps;
                if (isDef(effect.stride))
                    ssrPass.stride = effect.stride;
                if (isDef(effect.binarySearchSteps))
                    ssrPass.binarySearchSteps = effect.binarySearchSteps;
                if (isDef(effect.renderTargetScale))
                    ssrPass.renderTargetScale = effect.renderTargetScale;
                if (isDef(effect.thickness))
                    ssrPass.thickness = effect.thickness;
                if (isDef(effect.maxDistance))
                    ssrPass.maxDistance = effect.maxDistance;
                if (isDef(effect.jitter))
                    ssrPass.jitter = effect.jitter;
                if (isDef(effect.renderAfter))
                    ssrPass.renderAfter = effect.renderAfter;
                if (isDef(effect.simpleRefraction))
                    ssrPass.simpleRefraction = effect.simpleRefraction;

                // NOTE: temporary
                if (this.onResize) this.onResize();

                break;
            default:
                console.error('v3d.App: wrong postprocessing effect');
                break;
            }
        }

        if (composer.passes.length == 1) {
            var copyPass = new ShaderPass(ShaderLib.copy);
            composer.passes.push(copyPass);
            this.postprocessing.copyPass = copyPass;
        }

    },

    disablePostprocessing: function(keepOutline) {

        if (!this.postprocessing)
            return;

        var passesToRemove = ['bloomPass',
                              'brightnessContrastPass',
                              'bokehPass',
                              'grayscalePass',
                              'ssaoPass',
                              'ssrPassReflect',
                              'ssrPassRefract'];

        if (!keepOutline)
            passesToRemove.push('outlinePass');

        var composer = this.postprocessing.composer;

        for (var i = 0; i < passesToRemove.length; i++) {
            var passName = passesToRemove[i];

            if (this.postprocessing[passName]) {
                composer.passes.splice(composer.passes.indexOf(
                    this.postprocessing[passName]), 1);

                this.postprocessing[passName].dispose();
                this.postprocessing[passName] = null;
            }
        }

        // cleanup everything if only Render (and FXAA) remain
        if (!this.useHDR && this.aaMethod == 'AUTO' && composer.passes.length <= 2) {

            this.postprocessing.renderPass.dispose();

            if (this.postprocessing.fxaaPass)
                this.postprocessing.fxaaPass.dispose();

            this.postprocessing.composer.dispose();

            this.postprocessing = null;

        }

    },

    enableSSAA: function(sampleLevel, iterative) {
        if (!this.postprocessing)
            this.enablePostprocessing([]);

        // SSAA works bad with SSR since we have no depth info  and objects are rendered after main pass
        if (sampleLevel > 0 && (this.postprocessing.ssrPassRefract || this.postprocessing.ssrPassReflect))
            return;

        var composer = this.postprocessing.composer;

        if (sampleLevel > 0 && composer.passes[0] instanceof RenderPass) {

            if (!this.postprocessing.ssaaRenderPass) {
                var ssaaRenderPass = new SSAARenderPass(this.scene, this.camera);
                this.postprocessing.ssaaRenderPass = ssaaRenderPass;
            } else {
                var ssaaRenderPass = this.postprocessing.ssaaRenderPass;
            }

            // 16x
            ssaaRenderPass.sampleLevel = sampleLevel || 4;
            if (iterative) {
                ssaaRenderPass.iterative = true;
                ssaaRenderPass.iterativeFrame = 0;
            }

            composer.passes[0] = ssaaRenderPass;

            // disable FXAA if any
            if (this.postprocessing.fxaaPass)
                composer.passes.splice(composer.passes.length-1, 1);

        } else if (sampleLevel > 0) {

            composer.passes[0].sampleLevel = sampleLevel || 4;
            if (iterative) {
                ssaaRenderPass.iterative = true;
                ssaaRenderPass.iterativeFrame = 0;
            }

        } else if (sampleLevel == 0 && composer.passes[0] instanceof SSAARenderPass) {
            composer.passes[0] = this.postprocessing.renderPass;

            if (this.postprocessing.fxaaPass)
                composer.passes.push(this.postprocessing.fxaaPass);
        }
    },

    updateEnvironment: function(wMat) {

        this.disposeEnvironment();

        // NOTE: unconditionally
        wMat.useHDR = true;

        var cubeRTWorld = RenderUtils.renderWorldNodeMatToCubemap(
                this.renderer, wMat, this.worldCubemapRes);

        if (wMat.hasNode('LIGHT_PATH') || wMat.hasNode('AI_SKYDOME_LIGHT')) {
            var oldVal = wMat.defines['LIGHT_PATH_IS_CAM_RAY'];
            wMat.defines['LIGHT_PATH_IS_CAM_RAY'] = 0;
            var cubeRTEnv = RenderUtils.renderWorldNodeMatToCubemap(
                    this.renderer, wMat, this.worldCubemapRes);
            wMat.defines['LIGHT_PATH_IS_CAM_RAY'] = oldVal;

            this.initPMREM(cubeRTEnv);
            cubeRTEnv.dispose();
        } else {
            this.initPMREM(cubeRTWorld);
        }

        this.scene.background = new SceneBackground(cubeRTWorld);
        this.scene.background.useHDR = this.useHDR;
    },

    disposeEnvironment: function() {
        if (this._cubeUVEnvRT) {
            this._cubeUVEnvRT.dispose();
            this._traverseSceneForEnvUpdates(this.scene, function(obj) {
                obj.material.envMap.dispose();
            });
        }
        this._cubeUVEnvRT = null;

        var bkg = this.scene.background;
        if (bkg) {
            if (bkg.isSceneBackground) bkg = bkg.data;
            if (bkg && (bkg.isTexture || bkg.isWebGLRenderTarget)) {
                bkg.dispose();
            }
        }
    },

    initPMREM: function(cubeRT) {

        // process WebGLRenderTargetCube only
        if (!cubeRT.isWebGLRenderTargetCube)
            return;

        var pmremGenerator = new PMREMGenerator(cubeRT.texture);
        // don't flip WebGLRenderTargetCube textures along the X coordinate
        pmremGenerator.flipX = false;
        pmremGenerator.update(this.renderer);

        var pmremCubeUVPacker = new PMREMCubeUVPacker(pmremGenerator.cubeLods);
        pmremCubeUVPacker.update(this.renderer);

        this._cubeUVEnvRT = pmremCubeUVPacker.CubeUVRenderTarget;
        this._updateSceneCubeUVEnv(this.scene);

        pmremGenerator.dispose();
        pmremCubeUVPacker.dispose();
    },

    initWebXR: function(mode, referenceSpaceType, successCb, failureCb, exitCb) {

        successCb = successCb || function() {};
        failureCb = failureCb || function() {};
        exitCb = exitCb || function() {};

        var scope = this;

        function onSessionEnded(event) {

            scope.xrSession.removeEventListener('end', onSessionEnded);
            scope.renderer.vr.setSession(null);

            // allow unregister controllers first
            exitCb();

            if (mode == 'immersive-vr') {
                for (var i = 0; i < scope.xrControllers.length; i++) {
                    var controller = scope.xrControllers[i];

                    for (var j = controller.children.length-1; j >= 0; j--) {
                        var child = controller.children[j];

                        child.geometry.dispose();
                        child.material.dispose();

                        controller.remove(child);
                    }

                    scope.scene.remove(controller);

                }

                scope.xrControllers = [];
            }

            scope.xrSession = null;

            // return postprocessing back
            scope.postprocessing = scope._postprocessingSave;
            scope._postprocessingSave = null;

            scope.onResize();
        }

        navigator.xr.requestSession(mode).then(function(session) {

            scope._postprocessingSave = scope.postprocessing;
            scope.postprocessing = null;

            scope.renderer.vr.setReferenceSpaceType(referenceSpaceType);

            scope.xrSession = session;
            session.addEventListener('end', onSessionEnded);

            if (mode == 'immersive-vr') {

                // NOTE: reduce pixel ratio for VR sessions on HIDPI screens
                if (window.devicePixelRatio >= 2)
                    scope.renderer.vr.setFramebufferScaleFactor(0.5);

                var geometry = new BufferGeometry().setFromPoints([new Vector3(0, 0, 0),
                        new Vector3(0, 0, -1)]);

                // TODO:
                var numControllers = 1;

                for (var i = 0; i < numControllers; i++) {
                    var controller = scope.renderer.vr.getController(i);
                    scope.scene.add(controller);
                    scope.xrControllers.push(controller);

                    var line = new Line(geometry);
                    line.name = 'XR_CONTROLLER_RAY';
                    line.scale.z = 5;

                    controller.add(line.clone());

                    var spriteMap = new v3d.TextureLoader().load(VRReticleImage);
                    var spriteMaterial = new v3d.SpriteMaterial({ map: spriteMap, sizeAttenuation: false });
                    var sprite = new v3d.Sprite(spriteMaterial);
                    sprite.name = 'XR_CONTROLLER_RETICLE';
                    sprite.scale.multiplyScalar(0.1);
                    sprite.position.z = -5;
                    controller.add(sprite);
                }

            } else {
                // screen controller
                var controller = scope.renderer.vr.getController(0);
                scope.scene.add(controller);
                scope.xrControllers.push(controller);

                scope.scene.background = null;
                scope.renderer.autoClear = false;
                scope.onResize();
            }

            scope.renderer.vr.setImmersive(mode == 'immersive-vr' ? true : false);
            scope.renderer.vr.enabled = true;
            scope.renderer.vr.setSession(session);

            successCb();

        }).catch(failureCb);

    },

    endWebXR: function() {
        if (this.xrSession)
            this.xrSession.end();
    },

    printPerformanceInfo: function(delta) {
        var scope = this;

        delta = delta || 1;

        var renderer = scope.renderer;

        var info = renderer.info;

        var autoResetSave = info.autoReset;
        info.autoReset = false;
        info.reset();

        var frameStart = scope.frame;

        for (var i = 0; i < info.programs.length; i++) {
            info.programs[i].profile.calcRenderTime = true;
        }

        setTimeout(function() {

            console.log('--- Verge3D Performance Profile (' + delta + 's) ---');

            var frames = scope.frame - frameStart;

            console.log('Scene Loading Time:', String(Math.round(scope._loadingTime)) + 'ms');

            console.log('FPS:', frames / delta);
            console.log('Render Calls:', info.render.calls / frames);
            console.log('Triangles Rendered:', info.render.triangles / frames);

            console.log('Geometry Buffers:', info.memory.geometries);
            console.log('Textures & Render Buffers:', info.memory.textures);

            console.log('Materials and Shaders:');

            var programsSorted = info.programs.slice(0);

            programsSorted.sort(function(a,b) {
                return b.profile.renderTime - a.profile.renderTime;
            });

            var totalRenderTime = 0;

            for (var i = 0; i < programsSorted.length; i++) {
                var program = programsSorted[i];
                program.profile.calcRenderTime = false;

                var materials = program.profile.materials;

                var renderTime = program.profile.renderTime;

                console.log('   ', materials.length == 0 ? 'System Material' : materials.join(','),
                        '-', program.name, '-', String(Math.round(renderTime)) + 'ms');

                totalRenderTime += renderTime;
                program.profile.renderTime = 0;
            }

            console.log('Total Render Time:', String(Math.round(totalRenderTime)) + 'ms');
            //console.log('Real FPS:', Math.round(1000 / (totalRenderTime / frames)));

            info.autoReset = autoResetSave;

        }, delta * 1000);

    }

});

export { App };
